# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""MLIR lowering for enum operations."""

import operator

import numpy as np
from numba_cuda_mlir._mlir.dialects import arith

from numba_cuda_mlir import types
from numba_cuda_mlir.lowering_registry import LoweringRegistry
from numba_cuda_mlir.lowering_utilities import convert
from numba_cuda_mlir.lowering_utilities.type_conversions import to_mlir_type

registry = LoweringRegistry()
lower = registry.lower
lower_getattr = registry.lower_getattr
lower_cast = registry.lower_cast


def _is_float_enum(enum_type):
    return isinstance(enum_type.dtype, types.Float)


def _enum_cmp(builder, target, args, int_pred, float_pred, different_type_result):
    """Compare two EnumMembers. Different types return constant result."""
    lhs, rhs = args
    lhs_type = builder.get_numba_type(lhs.name)
    rhs_type = builder.get_numba_type(rhs.name)
    if lhs_type == rhs_type:
        lhs_val = builder.load_var(lhs)
        rhs_val = builder.load_var(rhs)
        if _is_float_enum(lhs_type):
            if float_pred is None:
                raise ValueError(
                    f"float comparison predicate required for float-backed enum type {lhs_type}"
                )
            result = arith.cmpf(float_pred, lhs_val, rhs_val)
        else:
            result = arith.cmpi(int_pred, lhs_val, rhs_val)
    else:
        result = arith.constant(arith.IntegerType.get_signless(1), different_type_result)
    builder.store_var(target, result)


@lower(operator.eq, types.EnumMember, types.EnumMember)
@lower(operator.is_, types.EnumMember, types.EnumMember)
def enum_eq(builder, target, args, kws):
    _enum_cmp(builder, target, args, arith.CmpIPredicate.eq, arith.CmpFPredicate.OEQ, 0)


@lower(operator.ne, types.EnumMember, types.EnumMember)
@lower(operator.is_not, types.EnumMember, types.EnumMember)
def enum_ne(builder, target, args, kws):
    _enum_cmp(builder, target, args, arith.CmpIPredicate.ne, arith.CmpFPredicate.UNE, 1)


@lower_getattr(types.EnumMember, "value")
@lower_getattr(types.IntEnumMember, "value")
def enum_value(context, builder, target, value):
    val = builder.load_var(value)
    builder.store_var(target, val)


@lower_cast(types.IntEnumMember, types.Integer)
def int_enum_to_int(builder, target, value):
    val = builder.load_var(value)
    target_type = builder.get_numba_type(target.name)
    result = convert(val, to_mlir_type(target_type))
    builder.store_var(target, result)


def _int_enum_value_cmp(builder, target, args, predicate):
    """IntEnum comparison by value (not type)."""
    lhs, rhs = args
    lhs_val = convert(builder.load_var(lhs), builder.load_var(rhs).type)
    rhs_val = builder.load_var(rhs)
    result = arith.cmpi(predicate, lhs_val, rhs_val)
    builder.store_var(target, result)


@lower(operator.eq, types.IntEnumMember, types.IntEnumMember)
def int_enum_eq(builder, target, args, kws):
    _int_enum_value_cmp(builder, target, args, arith.CmpIPredicate.eq)


@lower(operator.ne, types.IntEnumMember, types.IntEnumMember)
def int_enum_ne(builder, target, args, kws):
    _int_enum_value_cmp(builder, target, args, arith.CmpIPredicate.ne)


@lower(operator.is_, types.IntEnumMember, types.IntEnumMember)
def int_enum_is(builder, target, args, kws):
    _enum_cmp(builder, target, args, arith.CmpIPredicate.eq, None, 0)


@lower(operator.is_not, types.IntEnumMember, types.IntEnumMember)
def int_enum_is_not(builder, target, args, kws):
    _enum_cmp(builder, target, args, arith.CmpIPredicate.ne, None, 1)


@lower("static_getitem", types.EnumClass, types.StringLiteral)
@lower(operator.getitem, types.EnumClass, types.StringLiteral)
def enum_class_getitem(builder, target, args, kwargs):
    """Return an enum member by string key: Color["red"]."""
    enum_class_var, key_var = args
    enum_class_type = builder.get_numba_type(enum_class_var.name)
    key = builder.get_numba_type(key_var.name).literal_value
    member = enum_class_type.instance_class[key]
    mlir_type = to_mlir_type(enum_class_type.dtype)
    builder.store_var(target, arith.constant(mlir_type, member.value))


def _int_enum_mixed_values(builder, args):
    """Load and convert int/IntEnum operands to compatible types."""
    lhs, rhs = args
    lhs_val = builder.load_var(lhs)
    rhs_val = builder.load_var(rhs)
    lhs_type = builder.get_numba_type(lhs.name)
    rhs_type = builder.get_numba_type(rhs.name)
    if isinstance(lhs_type, types.IntEnumMember):
        lhs_val = convert(lhs_val, rhs_val.type)
    elif isinstance(rhs_type, types.IntEnumMember):
        rhs_val = convert(rhs_val, lhs_val.type)
    return lhs_val, rhs_val


@lower(operator.gt, types.Integer, types.IntEnumMember)
@lower(operator.gt, types.IntEnumMember, types.Integer)
def int_enum_gt(builder, target, args, kws):
    lhs, rhs = _int_enum_mixed_values(builder, args)
    builder.store_var(target, arith.cmpi(arith.CmpIPredicate.sgt, lhs, rhs))


@lower(operator.lt, types.Integer, types.IntEnumMember)
@lower(operator.lt, types.IntEnumMember, types.Integer)
def int_enum_lt(builder, target, args, kws):
    lhs, rhs = _int_enum_mixed_values(builder, args)
    builder.store_var(target, arith.cmpi(arith.CmpIPredicate.slt, lhs, rhs))


@lower(operator.ge, types.Integer, types.IntEnumMember)
@lower(operator.ge, types.IntEnumMember, types.Integer)
def int_enum_ge(builder, target, args, kws):
    lhs, rhs = _int_enum_mixed_values(builder, args)
    builder.store_var(target, arith.cmpi(arith.CmpIPredicate.sge, lhs, rhs))


@lower(operator.le, types.Integer, types.IntEnumMember)
@lower(operator.le, types.IntEnumMember, types.Integer)
def int_enum_le(builder, target, args, kws):
    lhs, rhs = _int_enum_mixed_values(builder, args)
    builder.store_var(target, arith.cmpi(arith.CmpIPredicate.sle, lhs, rhs))


@lower(operator.add, types.Integer, types.IntEnumMember)
@lower(operator.add, types.IntEnumMember, types.Integer)
def int_enum_add(builder, target, args, kws):
    lhs, rhs = _int_enum_mixed_values(builder, args)
    builder.store_var(target, arith.addi(lhs, rhs))


@lower(operator.sub, types.Integer, types.IntEnumMember)
@lower(operator.sub, types.IntEnumMember, types.Integer)
def int_enum_sub(builder, target, args, kws):
    lhs, rhs = _int_enum_mixed_values(builder, args)
    builder.store_var(target, arith.subi(lhs, rhs))


def _enum_to_number(builder, target, args, kws):
    """Convert IntEnumMember or EnumMember to numeric type: int(x), float(x), etc."""
    val = builder.load_var(args[0])
    target_type = builder.get_numba_type(target.name)
    builder.store_var(target, convert(val, to_mlir_type(target_type)))


_numeric_types = [
    int,
    float,
    types.int8,
    types.int16,
    types.int32,
    types.int64,
    types.uint8,
    types.uint16,
    types.uint32,
    types.uint64,
    types.float32,
    types.float64,
    np.int8,
    np.int16,
    np.int32,
    np.int64,
    np.uint8,
    np.uint16,
    np.uint32,
    np.uint64,
    np.float32,
    np.float64,
]
for typ in _numeric_types:
    lower(typ, types.IntEnumMember)(_enum_to_number)
    lower(typ, types.EnumMember)(_enum_to_number)



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""Utilities for materializing string constants as MLIR values.

String literals are kept as Python ``str`` throughout lowering so that
compile-time operations (e.g. constant-folded equality) can work on
the raw value.  Call ``materialize_string_constant_if_needed`` at the
boundary where an ``ir.Value`` is actually required (e.g. lowerings
that accept both ``UnicodeType`` and ``StringLiteral`` operands).
"""

import sys

from numba_cuda_mlir._mlir.extras import types as T
from numba_cuda_mlir._mlir import ir
from numba_cuda_mlir.mlir.dialect_exts import llvm
from numba_cuda_mlir.lowering_utilities import constant
from numba_cuda_mlir._threading import _LockedCounter

_HASH_WIDTH = 64 if sys.maxsize > 2**32 else 32
_counter = _LockedCounter(1)


def materialize_string_constant_if_needed(gpu_module, val):
    """If *val* is a Python str, materialize it as an MLIR unicode struct.

    Returns *val* unchanged if it is already an ``ir.Value``.
    """
    if isinstance(val, str):
        return materialize_string_constant(gpu_module, val)
    return val


def materialize_string_constant(gpu_module, pystr: str) -> ir.Value:
    """Materialize a Python string as an MLIR unicode struct SSA value.

    Emits an llvm.mlir.global with the string's byte data in *gpu_module*,
    then builds a unicode_type struct pointing at it.  The meminfo and
    parent fields are NULL (constant strings are not refcounted).

    Returns an SSA value of type
    ``!llvm.struct<(ptr, i64, i32, i32, i64, ptr, ptr)>``.
    """
    from numba_cuda_mlir.numba_cuda.cpython.unicode import compile_time_get_string_data

    databytes, length, kind, is_ascii, _ = compile_time_get_string_data(pystr)

    global_name = f"__numba_cuda_mlir_str_{next(_counter)}"

    arr_type = ir.Type.parse(f"!llvm.array<{len(databytes)} x i8>")
    gpu_block = gpu_module.bodyRegion.blocks[0]
    with ir.InsertionPoint.at_block_begin(gpu_block):
        linkage = ir.Attribute.parse("#llvm.linkage<internal>")
        str_value = ir.StringAttr.get(bytes(databytes).decode("latin-1"))
        llvm.GlobalOp(
            arr_type,
            global_name,
            linkage,
            addr_space=0,
            constant=True,
            value=str_value,
        )

    data_ptr = llvm.addressof(global_name)

    hash_type = ir.IntegerType.get_signless(_HASH_WIDTH)
    struct_type = llvm.StructType.get_literal(
        [llvm.ptr(), T.i64(), T.i32(), T.i32(), hash_type, llvm.ptr(), llvm.ptr()]
    )
    desc = llvm.mlir_undef(res=struct_type)
    desc = llvm.insertvalue(desc, data_ptr, 0)
    desc = llvm.insertvalue(desc, constant(length, T.i64()), 1)
    desc = llvm.insertvalue(desc, constant(kind, T.i32()), 2)
    desc = llvm.insertvalue(desc, constant(is_ascii, T.i32()), 3)
    desc = llvm.insertvalue(desc, constant(-1, hash_type), 4)
    desc = llvm.insertvalue(desc, llvm.mlir_zero(res=llvm.ptr()), 5)
    desc = llvm.insertvalue(desc, llvm.mlir_zero(res=llvm.ptr()), 6)
    return desc



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from .intrin import *
from .libdevice import *
from .jit import *

def syncthreads() -> None: ...
def syncwarp() -> None: ...
def consteval(value: Any) -> Any:
    """
    Force evaluation of an expression at compile time.
    """

def literal_unroll(value: Any) -> Any:
    """
    Unroll a loop at compile time.

    For compile-time range unrolling, use ``consteval`` from
    ``cuda.experimental``:

        >>> from numba_cuda_mlir.cuda.experimental import consteval
        >>> for i in consteval(range(10)):
        ...     array[i] = i
    """



#  Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
#  See https://llvm.org/LICENSE.txt for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from .._tensor_transform_ops_gen import *
from .._tensor_transform_ops_gen import _Dialect

try:
    from ...ir import *
    from ...dialects import transform
    from .._ods_common import _cext as _ods_cext
except ImportError as e:
    raise RuntimeError("Error loading imports from extension module") from e

from typing import Optional, overload, Union


@_ods_cext.register_operation(_Dialect, replace=True)
class MakeLoopIndependentOp(MakeLoopIndependentOp):
    """Specialization for MakeLoopIndependentOp class."""

    @overload
    def __init__(
        self,
        transformed_type: Type,
        target: Union[Operation, OpView, Value],
        num_loops: Union[int, IntegerAttr],
        *,
        loc=None,
        ip=None,
    ):
        ...

    @overload
    def __init__(
        self,
        target: Union[Operation, OpView, Value],
        num_loops: Union[int, IntegerAttr],
        *,
        loc=None,
        ip=None,
    ):
        ...

    def __init__(
        self,
        transformed_type_or_target: Type,
        target_or_num_loops: Union[int, IntegerAttr, Operation, OpView, Value] = None,
        num_loops_or_none: Optional[Union[int, IntegerAttr]] = None,
        *,
        loc=None,
        ip=None,
    ):
        if isinstance(transformed_type_or_target, Type):
            transformed_type = transformed_type_or_target
            target = target_or_num_loops
            num_loops = num_loops_or_none
        else:
            transformed_type = transform.AnyOpType.get()
            target = transformed_type_or_target
            num_loops = target_or_num_loops

        super().__init__(
            transformed_type,
            target,
            num_loops,
            loc=loc,
            ip=ip,
        )



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from numba_cuda_mlir.cuda.misc.special import literal_unroll, literally

__all__ = ["literal_unroll", "literally"]



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""Lowering for half-precision (fp16/bf16) intrinsic functions."""

import operator
from numba_cuda_mlir.lowering_utilities import convert
from numba_cuda_mlir.lowering_registry import LoweringRegistry

registry = LoweringRegistry()
lower = registry.lower
from numba_cuda_mlir.numba_cuda.types.ext_types import bfloat16 as bf16
from numba_cuda_mlir.numba_cuda import types
from numba_cuda_mlir._mlir.extras import types as T
from numba_cuda_mlir._mlir.dialects import arith, math as math_dialect


def _lower_unary_intrinsic(func, mlir_op, numba_type):
    """Register lowering for a unary half-precision intrinsic."""

    @lower(func, numba_type)
    def impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = mlir_op(value)
        builder.store_var(target, result)


def _lower_binary_intrinsic(func, mlir_op, numba_type):
    """Register lowering for a binary half-precision intrinsic."""

    @lower(func, numba_type, numba_type)
    def impl(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = mlir_op(lhs, rhs)
        builder.store_var(target, result)


def _lower_ternary_intrinsic(func, mlir_op, numba_type):
    """Register lowering for a ternary half-precision intrinsic."""

    @lower(func, numba_type, numba_type, numba_type)
    def impl(builder, target, args, kwargs):
        a = builder.load_var(args[0])
        b = builder.load_var(args[1])
        c = builder.load_var(args[2])
        result = mlir_op(a, b, c)
        builder.store_var(target, result)


def register_bf16_lowering():
    """Register lowering for bfloat16 intrinsics from numba_cuda_mlir.numba_cuda._internal.cuda_bf16."""
    from numba_cuda_mlir.numba_cuda._internal.cuda_bf16 import (
        nv_bfloat16,
        htrunc,
        hceil,
        hfloor,
        hrint,
        hsqrt,
        hrsqrt,
        hrcp,
        hlog,
        hlog2,
        hlog10,
        hcos,
        hsin,
        hexp,
        hexp2,
        hexp10,
        htanh,
        htanh_approx,
        # Additional intrinsics
        __habs,
        __hneg,
        __hadd,
        __hsub,
        __hmul,
        __hdiv,
        __hadd_rn,
        __hsub_rn,
        __hmul_rn,
        __hadd_sat,
        __hsub_sat,
        __hmul_sat,
        __hfma,
        __hfma_sat,
        __hfma_relu,
        # Comparison intrinsics
        __heq,
        __hne,
        __hgt,
        __hlt,
        __hge,
        __hle,
        __hequ,
        __hneu,
        __hgtu,
        __hltu,
        __hgeu,
        __hleu,
        __hmax,
        __hmin,
        __hmax_nan,
        __hmin_nan,
        __hisinf,
        __hisnan,
        # Conversion intrinsics
        __float2bfloat16,
        __float2bfloat16_rn,
        __float2bfloat16_rz,
        __float2bfloat16_rd,
        __float2bfloat16_ru,
        __double2bfloat16,
        __bfloat162float,
        __short2bfloat16_rn,
        __short2bfloat16_rz,
        __short2bfloat16_rd,
        __short2bfloat16_ru,
        __bfloat162short_rn,
        __bfloat162short_rz,
        __bfloat162short_rd,
        __bfloat162short_ru,
        __ushort2bfloat16_rn,
        __ushort2bfloat16_rz,
        __ushort2bfloat16_rd,
        __ushort2bfloat16_ru,
        __bfloat162ushort_rn,
        __bfloat162ushort_rz,
        __bfloat162ushort_rd,
        __bfloat162ushort_ru,
        __int2bfloat16_rn,
        __int2bfloat16_rz,
        __int2bfloat16_rd,
        __int2bfloat16_ru,
        __bfloat162int_rn,
        __bfloat162int_rz,
        __bfloat162int_rd,
        __bfloat162int_ru,
        __uint2bfloat16_rn,
        __uint2bfloat16_rz,
        __uint2bfloat16_rd,
        __uint2bfloat16_ru,
        __bfloat162uint_rn,
        __bfloat162uint_rz,
        __bfloat162uint_rd,
        __bfloat162uint_ru,
        __ll2bfloat16_rn,
        __ll2bfloat16_rz,
        __ll2bfloat16_rd,
        __ll2bfloat16_ru,
        __bfloat162ll_rn,
        __bfloat162ll_rz,
        __bfloat162ll_rd,
        __bfloat162ll_ru,
        __ull2bfloat16_rn,
        __ull2bfloat16_rz,
        __ull2bfloat16_rd,
        __ull2bfloat16_ru,
        __bfloat162ull_rn,
        __bfloat162ull_rz,
        __bfloat162ull_rd,
        __bfloat162ull_ru,
        __bfloat162char_rz,
        __bfloat162uchar_rz,
        # Bitcast intrinsics
        __bfloat16_as_short,
        __bfloat16_as_ushort,
        __short_as_bfloat16,
        __ushort_as_bfloat16,
    )

    # Constructor: nv_bfloat16(float64)
    @lower(nv_bfloat16, types.float64)
    def bf16_ctor_f64(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = convert(value, T.bf16())
        builder.store_var(target, result)

    # Constructor: nv_bfloat16(float32)
    @lower(nv_bfloat16, types.float32)
    def bf16_ctor_f32(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = convert(value, T.bf16())
        builder.store_var(target, result)

    # Constructor: nv_bfloat16(float16)
    @lower(nv_bfloat16, types.float16)
    def bf16_ctor_f16(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = convert(value, T.bf16())
        builder.store_var(target, result)

    # Constructor: nv_bfloat16(int16)
    @lower(nv_bfloat16, types.int16)
    def bf16_ctor_i16(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = convert(value, T.bf16())
        builder.store_var(target, result)

    # Constructor: nv_bfloat16(int32)
    @lower(nv_bfloat16, types.int32)
    def bf16_ctor_i32(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = convert(value, T.bf16())
        builder.store_var(target, result)

    # Constructor: nv_bfloat16(int64)
    @lower(nv_bfloat16, types.int64)
    def bf16_ctor_i64(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = convert(value, T.bf16())
        builder.store_var(target, result)

    # Constructor: nv_bfloat16(uint16)
    @lower(nv_bfloat16, types.uint16)
    def bf16_ctor_u16(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = convert(value, T.bf16())
        builder.store_var(target, result)

    # Constructor: nv_bfloat16(uint32)
    @lower(nv_bfloat16, types.uint32)
    def bf16_ctor_u32(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = convert(value, T.bf16())
        builder.store_var(target, result)

    # Constructor: nv_bfloat16(uint64)
    @lower(nv_bfloat16, types.uint64)
    def bf16_ctor_u64(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = convert(value, T.bf16())
        builder.store_var(target, result)

    # Constructor: nv_bfloat16(IntegerLiteral) - handles literal ints like 5
    @lower(nv_bfloat16, types.IntegerLiteral)
    def bf16_ctor_intlit(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = convert(value, T.bf16())
        builder.store_var(target, result)

    # Constructor: nv_bfloat16(Literal) - handles other literals
    @lower(nv_bfloat16, types.Literal)
    def bf16_ctor_lit(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = convert(value, T.bf16())
        builder.store_var(target, result)

    # Binary operators: bf16 + bf16, bf16 - bf16, etc.
    @lower(operator.add, bf16, bf16)
    def bf16_add(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.addf(lhs, rhs)
        builder.store_var(target, result)

    @lower(operator.sub, bf16, bf16)
    def bf16_sub(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.subf(lhs, rhs)
        builder.store_var(target, result)

    @lower(operator.mul, bf16, bf16)
    def bf16_mul(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.mulf(lhs, rhs)
        builder.store_var(target, result)

    @lower(operator.truediv, bf16, bf16)
    def bf16_div(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.divf(lhs, rhs)
        builder.store_var(target, result)

    # Unary operators
    @lower(operator.neg, bf16)
    def bf16_neg(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.negf(value)
        builder.store_var(target, result)

    @lower(operator.pos, bf16)
    def bf16_pos(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        builder.store_var(target, value)

    # Comparison operators
    @lower(operator.eq, bf16, bf16)
    def bf16_eq(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.cmpf(arith.CmpFPredicate.OEQ, lhs, rhs)
        builder.store_var(target, result)

    @lower(operator.ne, bf16, bf16)
    def bf16_ne(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.cmpf(arith.CmpFPredicate.ONE, lhs, rhs)
        builder.store_var(target, result)

    @lower(operator.lt, bf16, bf16)
    def bf16_lt(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.cmpf(arith.CmpFPredicate.OLT, lhs, rhs)
        builder.store_var(target, result)

    @lower(operator.le, bf16, bf16)
    def bf16_le(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.cmpf(arith.CmpFPredicate.OLE, lhs, rhs)
        builder.store_var(target, result)

    @lower(operator.gt, bf16, bf16)
    def bf16_gt(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.cmpf(arith.CmpFPredicate.OGT, lhs, rhs)
        builder.store_var(target, result)

    @lower(operator.ge, bf16, bf16)
    def bf16_ge(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.cmpf(arith.CmpFPredicate.OGE, lhs, rhs)
        builder.store_var(target, result)

    # In-place operators
    @lower(operator.iadd, bf16, bf16)
    def bf16_iadd(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.addf(lhs, rhs)
        builder.store_var(target, result)

    @lower(operator.isub, bf16, bf16)
    def bf16_isub(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.subf(lhs, rhs)
        builder.store_var(target, result)

    @lower(operator.imul, bf16, bf16)
    def bf16_imul(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.mulf(lhs, rhs)
        builder.store_var(target, result)

    @lower(operator.itruediv, bf16, bf16)
    def bf16_itruediv(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.divf(lhs, rhs)
        builder.store_var(target, result)

    # Unary math operations
    _lower_unary_intrinsic(htrunc, math_dialect.trunc, bf16)
    _lower_unary_intrinsic(hceil, math_dialect.ceil, bf16)
    _lower_unary_intrinsic(hfloor, math_dialect.floor, bf16)
    _lower_unary_intrinsic(hrint, math_dialect.roundeven, bf16)
    _lower_unary_intrinsic(hsqrt, math_dialect.sqrt, bf16)
    _lower_unary_intrinsic(hrsqrt, math_dialect.rsqrt, bf16)
    _lower_unary_intrinsic(hlog, math_dialect.log, bf16)
    _lower_unary_intrinsic(hlog2, math_dialect.log2, bf16)
    _lower_unary_intrinsic(hlog10, math_dialect.log10, bf16)
    _lower_unary_intrinsic(hcos, math_dialect.cos, bf16)
    _lower_unary_intrinsic(hsin, math_dialect.sin, bf16)
    _lower_unary_intrinsic(hexp, math_dialect.exp, bf16)
    _lower_unary_intrinsic(hexp2, math_dialect.exp2, bf16)
    _lower_unary_intrinsic(htanh, math_dialect.tanh, bf16)
    _lower_unary_intrinsic(htanh_approx, math_dialect.tanh, bf16)

    # hrcp (reciprocal): 1/x
    @lower(hrcp, bf16)
    def hrcp_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        one = arith.constant(result=T.bf16(), value=1.0)
        result = arith.divf(one, value)
        builder.store_var(target, result)

    # hexp10: 10^x = exp(x * ln(10))
    @lower(hexp10, bf16)
    def hexp10_impl(builder, target, args, kwargs):
        import math

        value = builder.load_var(args[0])
        ln10 = arith.constant(result=T.bf16(), value=math.log(10))
        scaled = arith.mulf(value, ln10)
        result = math_dialect.exp(scaled)
        builder.store_var(target, result)

    # __habs: absolute value
    @lower(__habs, bf16)
    def habs_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = math_dialect.absf(value)
        builder.store_var(target, result)

    # __hneg: negation
    @lower(__hneg, bf16)
    def hneg_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.negf(value)
        builder.store_var(target, result)

    # Binary arithmetic intrinsics
    @lower(__hadd, bf16, bf16)
    def hadd_impl(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.addf(lhs, rhs)
        builder.store_var(target, result)

    @lower(__hsub, bf16, bf16)
    def hsub_impl(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.subf(lhs, rhs)
        builder.store_var(target, result)

    @lower(__hmul, bf16, bf16)
    def hmul_impl(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.mulf(lhs, rhs)
        builder.store_var(target, result)

    @lower(__hdiv, bf16, bf16)
    def hdiv_impl(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.divf(lhs, rhs)
        builder.store_var(target, result)

    # Round-to-nearest variants (same as regular for now)
    @lower(__hadd_rn, bf16, bf16)
    def hadd_rn_impl(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.addf(lhs, rhs)
        builder.store_var(target, result)

    @lower(__hsub_rn, bf16, bf16)
    def hsub_rn_impl(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.subf(lhs, rhs)
        builder.store_var(target, result)

    @lower(__hmul_rn, bf16, bf16)
    def hmul_rn_impl(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.mulf(lhs, rhs)
        builder.store_var(target, result)

    # Saturating operations: clamp result to [0.0, 1.0]
    def saturate_bf16(value):
        zero = arith.constant(result=T.bf16(), value=0.0)
        one = arith.constant(result=T.bf16(), value=1.0)
        clamped_low = arith.maximumf(value, zero)
        return arith.minimumf(clamped_low, one)

    @lower(__hadd_sat, bf16, bf16)
    def hadd_sat_impl(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        sum_val = arith.addf(lhs, rhs)
        result = saturate_bf16(sum_val)
        builder.store_var(target, result)

    @lower(__hsub_sat, bf16, bf16)
    def hsub_sat_impl(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        diff = arith.subf(lhs, rhs)
        result = saturate_bf16(diff)
        builder.store_var(target, result)

    @lower(__hmul_sat, bf16, bf16)
    def hmul_sat_impl(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        prod = arith.mulf(lhs, rhs)
        result = saturate_bf16(prod)
        builder.store_var(target, result)

    # FMA operations
    @lower(__hfma, bf16, bf16, bf16)
    def hfma_impl(builder, target, args, kwargs):
        a = builder.load_var(args[0])
        b = builder.load_var(args[1])
        c = builder.load_var(args[2])
        result = math_dialect.fma(a, b, c)
        builder.store_var(target, result)

    @lower(__hfma_sat, bf16, bf16, bf16)
    def hfma_sat_impl(builder, target, args, kwargs):
        a = builder.load_var(args[0])
        b = builder.load_var(args[1])
        c = builder.load_var(args[2])
        fma_result = math_dialect.fma(a, b, c)
        result = saturate_bf16(fma_result)
        builder.store_var(target, result)

    @lower(__hfma_relu, bf16, bf16, bf16)
    def hfma_relu_impl(builder, target, args, kwargs):
        a = builder.load_var(args[0])
        b = builder.load_var(args[1])
        c = builder.load_var(args[2])
        fma_result = math_dialect.fma(a, b, c)
        zero = arith.constant(result=T.bf16(), value=0.0)
        result = arith.maximumf(fma_result, zero)
        builder.store_var(target, result)

    # Comparison intrinsics (ordered)
    @lower(__heq, bf16, bf16)
    def heq_impl(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.cmpf(arith.CmpFPredicate.OEQ, lhs, rhs)
        builder.store_var(target, result)

    @lower(__hne, bf16, bf16)
    def hne_impl(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.cmpf(arith.CmpFPredicate.ONE, lhs, rhs)
        builder.store_var(target, result)

    @lower(__hgt, bf16, bf16)
    def hgt_impl(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.cmpf(arith.CmpFPredicate.OGT, lhs, rhs)
        builder.store_var(target, result)

    @lower(__hlt, bf16, bf16)
    def hlt_impl(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.cmpf(arith.CmpFPredicate.OLT, lhs, rhs)
        builder.store_var(target, result)

    @lower(__hge, bf16, bf16)
    def hge_impl(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.cmpf(arith.CmpFPredicate.OGE, lhs, rhs)
        builder.store_var(target, result)

    @lower(__hle, bf16, bf16)
    def hle_impl(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.cmpf(arith.CmpFPredicate.OLE, lhs, rhs)
        builder.store_var(target, result)

    # Comparison intrinsics (unordered - returns true if either operand is NaN)
    @lower(__hequ, bf16, bf16)
    def hequ_impl(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.cmpf(arith.CmpFPredicate.UEQ, lhs, rhs)
        builder.store_var(target, result)

    @lower(__hneu, bf16, bf16)
    def hneu_impl(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.cmpf(arith.CmpFPredicate.UNE, lhs, rhs)
        builder.store_var(target, result)

    @lower(__hgtu, bf16, bf16)
    def hgtu_impl(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.cmpf(arith.CmpFPredicate.UGT, lhs, rhs)
        builder.store_var(target, result)

    @lower(__hltu, bf16, bf16)
    def hltu_impl(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.cmpf(arith.CmpFPredicate.ULT, lhs, rhs)
        builder.store_var(target, result)

    @lower(__hgeu, bf16, bf16)
    def hgeu_impl(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.cmpf(arith.CmpFPredicate.UGE, lhs, rhs)
        builder.store_var(target, result)

    @lower(__hleu, bf16, bf16)
    def hleu_impl(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.cmpf(arith.CmpFPredicate.ULE, lhs, rhs)
        builder.store_var(target, result)

    # Min/max operations
    # CUDA __hmax returns the non-NaN operand (IEEE 754-2008 maxNum semantics)
    @lower(__hmax, bf16, bf16)
    def hmax_impl(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.maxnumf(lhs, rhs)
        builder.store_var(target, result)

    @lower(__hmin, bf16, bf16)
    def hmin_impl(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.minnumf(lhs, rhs)
        builder.store_var(target, result)

    # CUDA __hmax_nan propagates NaN (IEEE 754-2019 maximum semantics)
    @lower(__hmax_nan, bf16, bf16)
    def hmax_nan_impl(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.maximumf(lhs, rhs)
        builder.store_var(target, result)

    @lower(__hmin_nan, bf16, bf16)
    def hmin_nan_impl(builder, target, args, kwargs):
        lhs = builder.load_var(args[0])
        rhs = builder.load_var(args[1])
        result = arith.minimumf(lhs, rhs)
        builder.store_var(target, result)

    # Special value checks
    @lower(__hisinf, bf16)
    def hisinf_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        abs_val = math_dialect.absf(value)
        inf = arith.constant(result=T.bf16(), value=float("inf"))
        result = arith.cmpf(arith.CmpFPredicate.OEQ, abs_val, inf)
        builder.store_var(target, result)

    @lower(__hisnan, bf16)
    def hisnan_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.cmpf(arith.CmpFPredicate.UNO, value, value)
        builder.store_var(target, result)

    # Conversion intrinsics: float -> bf16
    @lower(__float2bfloat16, types.float32)
    def float2bfloat16_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.truncf(out=T.bf16(), in_=value)
        builder.store_var(target, result)

    @lower(__float2bfloat16_rn, types.float32)
    def float2bfloat16_rn_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.truncf(out=T.bf16(), in_=value)
        builder.store_var(target, result)

    @lower(__float2bfloat16_rz, types.float32)
    def float2bfloat16_rz_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.truncf(out=T.bf16(), in_=value)
        builder.store_var(target, result)

    @lower(__float2bfloat16_rd, types.float32)
    def float2bfloat16_rd_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.truncf(out=T.bf16(), in_=value)
        builder.store_var(target, result)

    @lower(__float2bfloat16_ru, types.float32)
    def float2bfloat16_ru_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.truncf(out=T.bf16(), in_=value)
        builder.store_var(target, result)

    @lower(__double2bfloat16, types.float64)
    def double2bfloat16_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.truncf(out=T.bf16(), in_=value)
        builder.store_var(target, result)

    @lower(__bfloat162float, bf16)
    def bfloat162float_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.extf(out=T.f32(), in_=value)
        builder.store_var(target, result)

    # Conversion intrinsics: short (int16) <-> bf16
    @lower(__short2bfloat16_rn, types.int16)
    def short2bfloat16_rn_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.sitofp(out=T.bf16(), in_=value)
        builder.store_var(target, result)

    @lower(__short2bfloat16_rz, types.int16)
    def short2bfloat16_rz_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.sitofp(out=T.bf16(), in_=value)
        builder.store_var(target, result)

    @lower(__short2bfloat16_rd, types.int16)
    def short2bfloat16_rd_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.sitofp(out=T.bf16(), in_=value)
        builder.store_var(target, result)

    @lower(__short2bfloat16_ru, types.int16)
    def short2bfloat16_ru_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.sitofp(out=T.bf16(), in_=value)
        builder.store_var(target, result)

    @lower(__bfloat162short_rn, bf16)
    def bfloat162short_rn_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        value = math_dialect.roundeven(value)
        result = arith.fptosi(out=T.i16(), in_=value)
        builder.store_var(target, result)

    @lower(__bfloat162short_rz, bf16)
    def bfloat162short_rz_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.fptosi(out=T.i16(), in_=value)
        builder.store_var(target, result)

    @lower(__bfloat162short_rd, bf16)
    def bfloat162short_rd_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        value = math_dialect.floor(value)
        result = arith.fptosi(out=T.i16(), in_=value)
        builder.store_var(target, result)

    @lower(__bfloat162short_ru, bf16)
    def bfloat162short_ru_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        value = math_dialect.ceil(value)
        result = arith.fptosi(out=T.i16(), in_=value)
        builder.store_var(target, result)

    # Conversion intrinsics: ushort (uint16) <-> bf16
    @lower(__ushort2bfloat16_rn, types.uint16)
    def ushort2bfloat16_rn_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.uitofp(out=T.bf16(), in_=value)
        builder.store_var(target, result)

    @lower(__ushort2bfloat16_rz, types.uint16)
    def ushort2bfloat16_rz_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.uitofp(out=T.bf16(), in_=value)
        builder.store_var(target, result)

    @lower(__ushort2bfloat16_rd, types.uint16)
    def ushort2bfloat16_rd_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.uitofp(out=T.bf16(), in_=value)
        builder.store_var(target, result)

    @lower(__ushort2bfloat16_ru, types.uint16)
    def ushort2bfloat16_ru_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.uitofp(out=T.bf16(), in_=value)
        builder.store_var(target, result)

    @lower(__bfloat162ushort_rn, bf16)
    def bfloat162ushort_rn_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        value = math_dialect.roundeven(value)
        result = arith.fptoui(out=T.i16(), in_=value)
        builder.store_var(target, result)

    @lower(__bfloat162ushort_rz, bf16)
    def bfloat162ushort_rz_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.fptoui(out=T.i16(), in_=value)
        builder.store_var(target, result)

    @lower(__bfloat162ushort_rd, bf16)
    def bfloat162ushort_rd_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        value = math_dialect.floor(value)
        result = arith.fptoui(out=T.i16(), in_=value)
        builder.store_var(target, result)

    @lower(__bfloat162ushort_ru, bf16)
    def bfloat162ushort_ru_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        value = math_dialect.ceil(value)
        result = arith.fptoui(out=T.i16(), in_=value)
        builder.store_var(target, result)

    # Conversion intrinsics: int (int32) <-> bf16
    @lower(__int2bfloat16_rn, types.int32)
    def int2bfloat16_rn_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.sitofp(out=T.bf16(), in_=value)
        builder.store_var(target, result)

    @lower(__int2bfloat16_rz, types.int32)
    def int2bfloat16_rz_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.sitofp(out=T.bf16(), in_=value)
        builder.store_var(target, result)

    @lower(__int2bfloat16_rd, types.int32)
    def int2bfloat16_rd_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.sitofp(out=T.bf16(), in_=value)
        builder.store_var(target, result)

    @lower(__int2bfloat16_ru, types.int32)
    def int2bfloat16_ru_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.sitofp(out=T.bf16(), in_=value)
        builder.store_var(target, result)

    @lower(__bfloat162int_rn, bf16)
    def bfloat162int_rn_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        value = math_dialect.roundeven(value)
        result = arith.fptosi(out=T.i32(), in_=value)
        builder.store_var(target, result)

    @lower(__bfloat162int_rz, bf16)
    def bfloat162int_rz_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.fptosi(out=T.i32(), in_=value)
        builder.store_var(target, result)

    @lower(__bfloat162int_rd, bf16)
    def bfloat162int_rd_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        value = math_dialect.floor(value)
        result = arith.fptosi(out=T.i32(), in_=value)
        builder.store_var(target, result)

    @lower(__bfloat162int_ru, bf16)
    def bfloat162int_ru_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        value = math_dialect.ceil(value)
        result = arith.fptosi(out=T.i32(), in_=value)
        builder.store_var(target, result)

    # Conversion intrinsics: uint (uint32) <-> bf16
    @lower(__uint2bfloat16_rn, types.uint32)
    def uint2bfloat16_rn_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.uitofp(out=T.bf16(), in_=value)
        builder.store_var(target, result)

    @lower(__uint2bfloat16_rz, types.uint32)
    def uint2bfloat16_rz_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.uitofp(out=T.bf16(), in_=value)
        builder.store_var(target, result)

    @lower(__uint2bfloat16_rd, types.uint32)
    def uint2bfloat16_rd_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.uitofp(out=T.bf16(), in_=value)
        builder.store_var(target, result)

    @lower(__uint2bfloat16_ru, types.uint32)
    def uint2bfloat16_ru_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.uitofp(out=T.bf16(), in_=value)
        builder.store_var(target, result)

    @lower(__bfloat162uint_rn, bf16)
    def bfloat162uint_rn_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        value = math_dialect.roundeven(value)
        result = arith.fptoui(out=T.i32(), in_=value)
        builder.store_var(target, result)

    @lower(__bfloat162uint_rz, bf16)
    def bfloat162uint_rz_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.fptoui(out=T.i32(), in_=value)
        builder.store_var(target, result)

    @lower(__bfloat162uint_rd, bf16)
    def bfloat162uint_rd_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        value = math_dialect.floor(value)
        result = arith.fptoui(out=T.i32(), in_=value)
        builder.store_var(target, result)

    @lower(__bfloat162uint_ru, bf16)
    def bfloat162uint_ru_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        value = math_dialect.ceil(value)
        result = arith.fptoui(out=T.i32(), in_=value)
        builder.store_var(target, result)

    # Conversion intrinsics: ll (int64) <-> bf16
    @lower(__ll2bfloat16_rn, types.int64)
    def ll2bfloat16_rn_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.sitofp(out=T.bf16(), in_=value)
        builder.store_var(target, result)

    @lower(__ll2bfloat16_rz, types.int64)
    def ll2bfloat16_rz_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.sitofp(out=T.bf16(), in_=value)
        builder.store_var(target, result)

    @lower(__ll2bfloat16_rd, types.int64)
    def ll2bfloat16_rd_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.sitofp(out=T.bf16(), in_=value)
        builder.store_var(target, result)

    @lower(__ll2bfloat16_ru, types.int64)
    def ll2bfloat16_ru_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.sitofp(out=T.bf16(), in_=value)
        builder.store_var(target, result)

    @lower(__bfloat162ll_rn, bf16)
    def bfloat162ll_rn_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        value = math_dialect.roundeven(value)
        result = arith.fptosi(out=T.i64(), in_=value)
        builder.store_var(target, result)

    @lower(__bfloat162ll_rz, bf16)
    def bfloat162ll_rz_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.fptosi(out=T.i64(), in_=value)
        builder.store_var(target, result)

    @lower(__bfloat162ll_rd, bf16)
    def bfloat162ll_rd_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        value = math_dialect.floor(value)
        result = arith.fptosi(out=T.i64(), in_=value)
        builder.store_var(target, result)

    @lower(__bfloat162ll_ru, bf16)
    def bfloat162ll_ru_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        value = math_dialect.ceil(value)
        result = arith.fptosi(out=T.i64(), in_=value)
        builder.store_var(target, result)

    # Conversion intrinsics: ull (uint64) <-> bf16
    @lower(__ull2bfloat16_rn, types.uint64)
    def ull2bfloat16_rn_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.uitofp(out=T.bf16(), in_=value)
        builder.store_var(target, result)

    @lower(__ull2bfloat16_rz, types.uint64)
    def ull2bfloat16_rz_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.uitofp(out=T.bf16(), in_=value)
        builder.store_var(target, result)

    @lower(__ull2bfloat16_rd, types.uint64)
    def ull2bfloat16_rd_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.uitofp(out=T.bf16(), in_=value)
        builder.store_var(target, result)

    @lower(__ull2bfloat16_ru, types.uint64)
    def ull2bfloat16_ru_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.uitofp(out=T.bf16(), in_=value)
        builder.store_var(target, result)

    @lower(__bfloat162ull_rn, bf16)
    def bfloat162ull_rn_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        value = math_dialect.roundeven(value)
        result = arith.fptoui(out=T.i64(), in_=value)
        builder.store_var(target, result)

    @lower(__bfloat162ull_rz, bf16)
    def bfloat162ull_rz_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.fptoui(out=T.i64(), in_=value)
        builder.store_var(target, result)

    @lower(__bfloat162ull_rd, bf16)
    def bfloat162ull_rd_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        value = math_dialect.floor(value)
        result = arith.fptoui(out=T.i64(), in_=value)
        builder.store_var(target, result)

    @lower(__bfloat162ull_ru, bf16)
    def bfloat162ull_ru_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        value = math_dialect.ceil(value)
        result = arith.fptoui(out=T.i64(), in_=value)
        builder.store_var(target, result)

    # Conversion intrinsics: char/uchar
    @lower(__bfloat162char_rz, bf16)
    def bfloat162char_rz_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.fptosi(out=T.i8(), in_=value)
        builder.store_var(target, result)

    @lower(__bfloat162uchar_rz, bf16)
    def bfloat162uchar_rz_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.fptoui(out=T.i8(), in_=value)
        builder.store_var(target, result)

    # Bitcast intrinsics: bf16 -> int16
    @lower(__bfloat16_as_short, bf16)
    def bfloat16_as_short_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.bitcast(T.i16(), value)
        builder.store_var(target, result)

    @lower(__bfloat16_as_ushort, bf16)
    def bfloat16_as_ushort_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.bitcast(T.i16(), value)
        builder.store_var(target, result)

    # Bitcast intrinsics: int16 -> bf16
    @lower(__short_as_bfloat16, types.int16)
    def short_as_bfloat16_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.bitcast(T.bf16(), value)
        builder.store_var(target, result)

    @lower(__ushort_as_bfloat16, types.uint16)
    def ushort_as_bfloat16_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.bitcast(T.bf16(), value)
        builder.store_var(target, result)

    # Bitcast intrinsics with int64/int32 args (for when scalars are typed as int64)
    # These truncate to i16 first, then treat that as the bf16 bit pattern
    @lower(__bfloat16_as_short, types.int64)
    def bfloat16_as_short_i64_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.trunci(T.i16(), value)
        builder.store_var(target, result)

    @lower(__bfloat16_as_short, types.int32)
    def bfloat16_as_short_i32_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.trunci(T.i16(), value)
        builder.store_var(target, result)

    @lower(__bfloat16_as_short, types.int16)
    def bfloat16_as_short_i16_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        builder.store_var(target, value)

    @lower(__bfloat16_as_ushort, types.int64)
    def bfloat16_as_ushort_i64_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.trunci(T.i16(), value)
        builder.store_var(target, result)

    @lower(__bfloat16_as_ushort, types.int32)
    def bfloat16_as_ushort_i32_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.trunci(T.i16(), value)
        builder.store_var(target, result)

    @lower(__bfloat16_as_ushort, types.int16)
    def bfloat16_as_ushort_i16_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        builder.store_var(target, value)

    @lower(__bfloat16_as_ushort, types.uint64)
    def bfloat16_as_ushort_u64_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.trunci(T.i16(), value)
        builder.store_var(target, result)

    @lower(__bfloat16_as_ushort, types.uint32)
    def bfloat16_as_ushort_u32_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        result = arith.trunci(T.i16(), value)
        builder.store_var(target, result)

    @lower(__bfloat16_as_ushort, types.uint16)
    def bfloat16_as_ushort_u16_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        builder.store_var(target, value)

    @lower(__short_as_bfloat16, types.int64)
    def short_as_bfloat16_i64_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        truncated = arith.trunci(T.i16(), value)
        result = arith.bitcast(T.bf16(), truncated)
        builder.store_var(target, result)

    @lower(__short_as_bfloat16, types.int32)
    def short_as_bfloat16_i32_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        truncated = arith.trunci(T.i16(), value)
        result = arith.bitcast(T.bf16(), truncated)
        builder.store_var(target, result)

    @lower(__ushort_as_bfloat16, types.int64)
    def ushort_as_bfloat16_i64_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        truncated = arith.trunci(T.i16(), value)
        result = arith.bitcast(T.bf16(), truncated)
        builder.store_var(target, result)

    @lower(__ushort_as_bfloat16, types.int32)
    def ushort_as_bfloat16_i32_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        truncated = arith.trunci(T.i16(), value)
        result = arith.bitcast(T.bf16(), truncated)
        builder.store_var(target, result)

    @lower(__ushort_as_bfloat16, types.uint64)
    def ushort_as_bfloat16_u64_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        truncated = arith.trunci(T.i16(), value)
        result = arith.bitcast(T.bf16(), truncated)
        builder.store_var(target, result)

    @lower(__ushort_as_bfloat16, types.uint32)
    def ushort_as_bfloat16_u32_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        truncated = arith.trunci(T.i16(), value)
        result = arith.bitcast(T.bf16(), truncated)
        builder.store_var(target, result)

    @lower(__short_as_bfloat16, types.IntegerLiteral)
    def short_as_bfloat16_lit_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        truncated = arith.trunci(T.i16(), value)
        result = arith.bitcast(T.bf16(), truncated)
        builder.store_var(target, result)


def register_fp16_lowering():
    """Register lowering for fp16 intrinsics from numba_cuda_mlir.cuda.fp16."""
    from numba_cuda_mlir.cuda import fp16

    # Unary operations missing from cuda.py
    _lower_unary_intrinsic(fp16.htrunc, math_dialect.trunc, types.float16)
    _lower_unary_intrinsic(fp16.hceil, math_dialect.ceil, types.float16)
    _lower_unary_intrinsic(fp16.hfloor, math_dialect.floor, types.float16)
    _lower_unary_intrinsic(fp16.hrint, math_dialect.roundeven, types.float16)
    _lower_unary_intrinsic(fp16.hsqrt, math_dialect.sqrt, types.float16)
    _lower_unary_intrinsic(fp16.hrsqrt, math_dialect.rsqrt, types.float16)
    _lower_unary_intrinsic(fp16.htanh, math_dialect.tanh, types.float16)
    _lower_unary_intrinsic(fp16.htanh_approx, math_dialect.tanh, types.float16)

    # hrcp (reciprocal): 1/x
    @lower(fp16.hrcp, types.float16)
    def fp16_hrcp_impl(builder, target, args, kwargs):
        value = builder.load_var(args[0])
        one = arith.constant(result=T.f16(), value=1.0)
        result = arith.divf(one, value)
        builder.store_var(target, result)

    # hexp10: 10^x = exp(x * ln(10))
    @lower(fp16.hexp10, types.float16)
    def fp16_hexp10_impl(builder, target, args, kwargs):
        import math

        value = builder.load_var(args[0])
        ln10 = arith.constant(result=T.f16(), value=math.log(10))
        scaled = arith.mulf(value, ln10)
        result = math_dialect.exp(scaled)
        builder.store_var(target, result)


register_bf16_lowering()
register_fp16_lowering()



#  Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
#  See https://llvm.org/LICENSE.txt for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from typing import (
    Dict,
    List,
    Union,
    Tuple,
    Any,
    Optional,
    Callable,
    TypeVar,
    get_origin,
    get_args,
)
from collections.abc import Sequence
from dataclasses import dataclass
from inspect import Parameter, Signature
from types import UnionType
from enum import Enum
from . import irdl
from ._ods_common import _cext, segmented_accessor
from .irdl import Variadicity
from ..passmanager import PassManager
from contextlib import nullcontext

ir = _cext.ir

__all__ = [
    "Dialect",
    "Operation",
    "Operand",
    "Result",
    "Region",
    "Type",
    "Attribute",
    "result",
    "infer_result",
    "operand",
    "attribute",
]

Operand = ir.Value
Result = ir.OpResult
Region = ir.Region


def construct_instance(origin, args):
    if not issubclass(origin, ir.Type | ir.Attribute):
        raise TypeError(f"unsupported type in constraints: {origin}")

    # `origin.get` is to construct an instance of MLIR type or attribute.
    return origin.get(
        *(
            (
                construct_instance(get_origin(arg), get_args(arg))
                if get_origin(arg)
                else arg
            )
            for arg in args
        )
    )


class ConstraintLoweringContext:
    def __init__(self):
        self._cache: Dict[str, ir.Value] = {}

    def lower(self, type_) -> ir.Value:
        """
        Lower a type hint (e.g. `Any`, `IntegerType[32]`, `IntegerAttr | StringAttr`) into IRDL ops.
        """

        if type(type_) is TypeVar:
            if type_.__name__ in self._cache:
                return self._cache[type_.__name__]
            v = self._lower(type_.__bound__ or Any)
            self._cache[type_.__name__] = v
        else:
            v = self._lower(type_)
        return v

    def _lower(self, type_) -> ir.Value:
        origin = get_origin(type_)
        if origin is UnionType or origin is Union:
            return irdl.any_of(self.lower(arg) for arg in get_args(type_))
        elif type_ is Any:
            return irdl.any()
        elif isinstance(type_, TypeVar):
            return self.lower(type_)
        elif origin and issubclass(origin, Type | Attribute):
            return irdl.parametric(
                base_type=[origin._dialect_name, origin._name],
                args=[self.lower(arg) for arg in get_args(type_)],
            )
        elif origin and issubclass(origin, ir.Type):
            t = construct_instance(origin, get_args(type_))
            return irdl.is_(ir.TypeAttr.get(t))
        elif origin and issubclass(origin, ir.Attribute):
            attr = construct_instance(origin, get_args(type_))
            return irdl.is_(attr)
        elif issubclass(type_, Type | Attribute):
            return irdl.base(base_ref=[type_._dialect_name, type_._name])
        elif issubclass(type_, ir.Type):
            return irdl.base(base_name=f"!{type_.type_name}")
        elif issubclass(type_, ir.Attribute):
            return irdl.base(base_name=f"#{type_.attr_name}")

        raise TypeError(f"unsupported type in constraints: {type_}")


@dataclass
class FieldSpecifier:
    type_: Any = None
    infer_type: bool = False
    default_is_none: bool = False
    default_factory: Optional[Callable[[], Any]] = None
    kw_only: bool = False

    @property
    def param_kind(self):
        if self.default_is_none or self.default_factory or self.infer_type:
            return ParameterKind.KEYWORD_ONLY_WITH_DEFAULT
        if self.kw_only:
            return ParameterKind.KEYWORD_ONLY_WITHOUT_DEFAULT
        return ParameterKind.POSITIONAL_OR_KEYWORD


def result(
    *,
    default_factory: Optional[Callable[[], Any]] = None,
    kw_only: bool = False,
) -> Result:
    """
    A field specifier for `Result` definitions.
    """

    return FieldSpecifier(
        type_=Result,
        default_factory=default_factory,
        kw_only=kw_only,
    )


def infer_result() -> Result:
    """
    A field specifier for `Result` definitions with type inference enabled.
    """

    return FieldSpecifier(
        type_=Result,
        infer_type=True,
    )


def operand(
    *,
    kw_only: bool = False,
) -> Operand:
    """
    A field specifier for `Operand` definitions.
    """

    return FieldSpecifier(
        type_=Operand,
        kw_only=kw_only,
    )


def attribute(
    *,
    default_factory: Optional[Callable[[], Any]] = None,
    kw_only: bool = False,
) -> ir.Attribute:
    """
    A field specifier for attribute definitions.
    """

    return FieldSpecifier(
        type_=Attribute,
        default_factory=default_factory,
        kw_only=kw_only,
    )


def infer_type_impl(type_) -> Callable[[], ir.Type]:
    """
    A function to infer ir.Type from type annotation.
    Returns a callable that returns the inferred ir.Type.
    We use callables so that MLIR contexts are not required
    while calling this function.
    """

    origin = get_origin(type_)
    if origin and issubclass(origin, ir.Type | ir.Attribute):
        args = [
            infer_type_impl(arg) if get_origin(arg) else lambda: arg
            for arg in get_args(type_)
        ]
        return lambda: origin.get(*[arg() for arg in args])
    elif isinstance(type_, TypeVar):
        return infer_type_impl(type_.__bound__)
    raise TypeError(f"unsupported type for inferring: {type_}")


class ParameterKind(Enum):
    POSITIONAL_OR_KEYWORD = 1
    KEYWORD_ONLY_WITHOUT_DEFAULT = 2
    KEYWORD_ONLY_WITH_DEFAULT = 3


@dataclass
class FieldDef:
    """
    Base class for kinds of fields that can occur in an `Operation`'s definition.
    """

    name: str
    variadicity: Variadicity
    constraint: Any

    param_kind: ParameterKind = ParameterKind.POSITIONAL_OR_KEYWORD

    @staticmethod
    def from_type_hint(name, type_, specifier) -> "FieldDef":
        variadicity = Variadicity.single
        if inner := match_optional(type_):
            variadicity = Variadicity.optional
            type_ = inner
        elif get_origin(type_) is Sequence:
            variadicity = Variadicity.variadic
            type_ = get_args(type_)[0]

        origin = get_origin(type_)
        if origin is ir.OpResult:
            if specifier.type_ and specifier.type_ is not Result:
                raise TypeError(
                    f"only `result` field specifier can be used for result fields"
                )
            constraint = get_args(type_)[0]
            return ResultDef(
                name,
                variadicity,
                constraint,
                param_kind=specifier.param_kind,
                default_factory=specifier.default_factory,
                default_is_none=specifier.default_is_none,
                infer_type=(
                    infer_type_impl(constraint) if specifier.infer_type else None
                ),
            )
        elif origin is ir.Value:
            if specifier.type_ and specifier.type_ is not Operand:
                raise TypeError(
                    f"only `operand` field specifier can be used for operand fields"
                )
            return OperandDef(
                name,
                variadicity,
                get_args(type_)[0],
                param_kind=specifier.param_kind,
                default_is_none=specifier.default_is_none,
            )
        elif type_ is ir.Region:
            if specifier.type_ and specifier.type_ is not Region:
                raise TypeError(
                    f"this field specifier can not be used for region fields"
                )
            return RegionDef(name, variadicity, Any)

        if specifier.type_ and specifier.type_ is not Attribute:
            raise TypeError(
                f"only `attribute` field specifier can be used for attribute fields"
            )
        return AttributeDef(
            name,
            variadicity,
            type_,
            param_kind=specifier.param_kind,
            default_factory=specifier.default_factory,
        )


@dataclass
class OperandDef(FieldDef):
    default_is_none: bool = False

    def __post_init__(self):
        if self.variadicity != Variadicity.optional and self.default_is_none:
            raise ValueError(f"only optional operand can be set to None")


@dataclass
class ResultDef(FieldDef):
    infer_type: Callable[[], ir.Type] | None = None
    default_factory: Optional[Callable[[], Any]] = None
    default_is_none: bool = False

    def __post_init__(self):
        if self.variadicity != Variadicity.optional and self.default_is_none:
            raise ValueError(f"only optional result can be set to None")

        if self.infer_type and self.variadicity != Variadicity.single:
            raise ValueError(
                f"type of variadic or optional result '{self.name}' cannot be inferred"
            )

    def process_type(self, type_):
        if type_:
            return type_

        if self.infer_type:
            return self.infer_type()

        if self.default_factory:
            return self.default_factory()

        return None


@dataclass
class AttributeDef(FieldDef):
    default_factory: Optional[Callable[[], Any]] = None

    def __post_init__(self):
        if self.variadicity != Variadicity.single:
            raise ValueError("optional attribute is not currently supported")
        if (
            self.param_kind == ParameterKind.KEYWORD_ONLY_WITH_DEFAULT
            and not self.default_factory
        ):
            raise ValueError(f"only optional attribute can be set to None")

    def process_attr(self, attr):
        if attr:
            return attr

        if self.default_factory:
            return self.default_factory()

        return None


@dataclass
class RegionDef(FieldDef):
    def __post_init__(self):
        if self.variadicity != Variadicity.single:
            raise ValueError("optional region is not currently supported")


def partition_fields(
    fields: List[FieldDef],
) -> Tuple[List[OperandDef], List[AttributeDef], List[ResultDef], List[RegionDef]]:
    operands = [i for i in fields if isinstance(i, OperandDef)]
    attrs = [i for i in fields if isinstance(i, AttributeDef)]
    results = [i for i in fields if isinstance(i, ResultDef)]
    regions = [i for i in fields if isinstance(i, RegionDef)]
    return operands, attrs, results, regions


def normalize_value_range(
    value_range: Union[ir.OpOperandList, ir.OpResultList],
    variadicity: Variadicity,
) -> ir.Value | ir.OpOperandList | ir.OpResultList | None:
    if variadicity == Variadicity.single:
        return value_range[0]
    if variadicity == Variadicity.optional:
        return value_range[0] if len(value_range) > 0 else None
    return value_range


def match_optional(type_) -> Optional[Any]:
    """
    Try to match type hint like `Optional[T]`, `T | None` or `None | T`.
    Returns the `T` inside `Optional[T]` if matched.
    Returns `None` if not matched.
    """

    origin = get_origin(type_)
    args = get_args(type_)
    if (
        (origin is Union or origin is UnionType)
        and len(args) == 2
        and type(None) in args
    ):
        return args[0] if args[1] is type(None) else args[1]

    return None


class Operation(ir.OpView):
    """
    Base class of Python-defined operations.

    The following example shows two ways to define operations via this class:
    ```python
    class MyOp(MyDialect.Operation, name=..):
      ...

    class MyOp(Operation, dialect=MyDialect, name=..):
      ...
    ```
    """

    def __init__(*args, **kwargs):
        raise TypeError(
            "This class is a template and cannot be instantiated directly. "
            "Please use a subclass that defines the operation."
        )

    @classmethod
    def __init_subclass__(
        cls,
        *,
        name: str | None = None,
        traits: list[type] | None = None,
        dialect: type | None = None,
        **kwargs,
    ):
        """
        This method is to perform all magic to make a `Operation` subclass works like a dataclass, like:
        - generate the method to emit IRDL operations,
        - generate `__init__` method as an operation builder function,
        - generate operand, result and attribute accessors
        """

        super().__init_subclass__(**kwargs)

        fields = []

        for base in cls.__bases__:
            if hasattr(base, "_fields"):
                fields.extend(base._fields)
        for key, value in cls.__annotations__.items():
            # if the class variable is not defined, we treat it as a default specifier;
            # if it is assigned with `None`, we treat it as a specifier with `default_is_none=True`.
            # e.g. x : int         # default specifier
            #      y : int = None  # specifier with default_is_none=True
            specifier = cls.__dict__.get(key, FieldSpecifier()) or FieldSpecifier(
                default_is_none=True
            )
            # treat all other values as invalid
            if not isinstance(specifier, FieldSpecifier):
                raise TypeError(
                    f"the field specifier of field '{key}' is not supported"
                )
            field = FieldDef.from_type_hint(key, value, specifier)
            fields.append(field)

        cls._fields = fields

        traits = traits or []

        for base in cls.__bases__:
            if hasattr(base, "_traits"):
                traits = base._traits + traits

        cls._traits = traits

        if dialect:
            if hasattr(cls, "_dialect_obj"):
                raise RuntimeError(
                    f"This operation has already been attached to dialect '{cls._dialect_obj.DIALECT_NAMESPACE}'."
                )
            cls._dialect_obj = dialect

        # for subclasses without "name" parameter,
        # just treat them as normal classes
        if not name:
            return

        if not hasattr(cls, "_dialect_obj"):
            raise RuntimeError(
                "Operation subclasses must either inherit from a Dialect's Operation subclass "
                "or provide the dialect as a class keyword argument."
            )

        op_name = name
        cls._op_name = op_name
        dialect_name = cls._dialect_obj.DIALECT_NAMESPACE
        dialect_obj = cls._dialect_obj

        cls._generate_class_attributes(dialect_name, op_name, fields)
        cls._generate_init_method(fields)
        operands, attrs, results, regions = partition_fields(fields)
        cls._generate_attr_properties(attrs)
        cls._generate_operand_properties(operands)
        cls._generate_result_properties(results)
        cls._generate_region_properties(regions)

        cls.Adaptor = type(
            "Adaptor",
            (OperationAdator,),
            dict(),
            operation=cls,
        )

        dialect_obj.operations.append(cls)

    @staticmethod
    def _variadicity_to_segment(variadicity: Variadicity) -> int:
        return {Variadicity.variadic: -1, Variadicity.optional: 0}.get(variadicity, 1)

    @staticmethod
    def _generate_segments(
        operands_or_results: List[Union[OperandDef, ResultDef]],
    ) -> List[int] | None:
        if any(i.variadicity != Variadicity.single for i in operands_or_results):
            return [
                Operation._variadicity_to_segment(i.variadicity)
                for i in operands_or_results
            ]
        return None

    @staticmethod
    def _generate_init_signature(fields: List[FieldDef]) -> Signature:
        args = [i for i in fields if not isinstance(i, RegionDef)]

        params = [Parameter("self", Parameter.POSITIONAL_ONLY)]

        for i in args:
            match i.param_kind:
                case ParameterKind.POSITIONAL_OR_KEYWORD:
                    params.append(Parameter(i.name, Parameter.POSITIONAL_OR_KEYWORD))
                case ParameterKind.KEYWORD_ONLY_WITH_DEFAULT:
                    params.append(
                        Parameter(i.name, Parameter.KEYWORD_ONLY, default=None)
                    )
                case ParameterKind.KEYWORD_ONLY_WITHOUT_DEFAULT:
                    params.append(Parameter(i.name, Parameter.KEYWORD_ONLY))

        params.append(Parameter("loc", Parameter.KEYWORD_ONLY, default=None))
        params.append(Parameter("ip", Parameter.KEYWORD_ONLY, default=None))

        return Signature(params)

    @classmethod
    def _generate_init_method(cls, fields: List[FieldDef]) -> None:
        operands, attrs, results, regions = partition_fields(fields)

        init_sig = cls._generate_init_signature(fields)

        def __init__(*args, **kwargs):
            bound = init_sig.bind(*args, **kwargs)
            bound.apply_defaults()
            args = bound.arguments

            _operands = [args[operand.name] for operand in operands]
            _results = [result.process_type(args[result.name]) for result in results]
            _attributes = dict(
                (attr.name, attr.process_attr(args[attr.name])) for attr in attrs
            )
            _regions = len(regions) or None
            _ods_successors = None
            self = args["self"]
            super(Operation, self).__init__(
                self.OPERATION_NAME,
                self._ODS_REGIONS,
                self._ODS_OPERAND_SEGMENTS,
                self._ODS_RESULT_SEGMENTS,
                attributes=_attributes,
                results=_results,
                operands=_operands,
                successors=_ods_successors,
                regions=_regions,
                loc=args["loc"],
                ip=args["ip"],
            )

        __init__.__signature__ = init_sig
        cls.__init__ = __init__

    @classmethod
    def _generate_class_attributes(
        cls, dialect_name: str, op_name: str, fields: List[FieldDef]
    ) -> None:
        operands, attrs, results, regions = partition_fields(fields)

        operand_segments = cls._generate_segments(operands)
        result_segments = cls._generate_segments(results)

        cls.OPERATION_NAME = f"{dialect_name}.{op_name}"
        cls._ODS_REGIONS = (len(regions), True)
        cls._ODS_OPERAND_SEGMENTS = operand_segments
        cls._ODS_RESULT_SEGMENTS = result_segments

    @classmethod
    def _generate_attr_properties(cls, attrs: List[AttributeDef]) -> None:
        for attr in attrs:
            setattr(
                cls,
                attr.name,
                property(lambda self, name=attr.name: self.attributes[name]),
            )

    @classmethod
    def _generate_region_properties(cls, regions: List[RegionDef]) -> None:
        for i, region in enumerate(regions):
            setattr(
                cls,
                region.name,
                property(lambda self, i=i: self.regions[i]),
            )

    @classmethod
    def _generate_operand_properties(cls, operands: List[OperandDef]) -> None:
        for i, operand in enumerate(operands):
            if cls._ODS_OPERAND_SEGMENTS:

                def getter(self, i=i, operand=operand):
                    operand_range = segmented_accessor(
                        self.operation.operands,
                        self.operation.attributes["operandSegmentSizes"],
                        i,
                    )
                    return normalize_value_range(operand_range, operand.variadicity)

                setattr(cls, operand.name, property(getter))
            else:
                setattr(cls, operand.name, property(lambda self, i=i: self.operands[i]))

    @classmethod
    def _generate_result_properties(cls, results: List[ResultDef]) -> None:
        for i, result in enumerate(results):
            if cls._ODS_RESULT_SEGMENTS:

                def getter(self, i=i, result=result):
                    result_range = segmented_accessor(
                        self.operation.results,
                        self.operation.attributes["resultSegmentSizes"],
                        i,
                    )
                    return normalize_value_range(result_range, result.variadicity)

                setattr(cls, result.name, property(getter))
            else:
                setattr(cls, result.name, property(lambda self, i=i: self.results[i]))

    @classmethod
    def _attach_traits(cls) -> None:
        for trait in cls._traits:
            trait.attach(cls.OPERATION_NAME)

        if hasattr(cls, "verify_invariants") or hasattr(
            cls, "verify_region_invariants"
        ):
            ir.DynamicOpTrait.attach(cls.OPERATION_NAME, cls)

    @classmethod
    def _emit_operation(cls) -> None:
        ctx = ConstraintLoweringContext()
        operands, attrs, results, regions = partition_fields(cls._fields)

        op = irdl.operation_(cls._op_name)
        with ir.InsertionPoint(op.body):
            if operands:
                irdl.operands_(
                    [ctx.lower(i.constraint) for i in operands],
                    [i.name for i in operands],
                    [i.variadicity for i in operands],
                )
            if attrs:
                irdl.attributes_(
                    [ctx.lower(i.constraint) for i in attrs],
                    [i.name for i in attrs],
                )
            if results:
                irdl.results_(
                    [ctx.lower(i.constraint) for i in results],
                    [i.name for i in results],
                    [i.variadicity for i in results],
                )
            if regions:
                irdl.regions_(
                    [irdl.region([]) for _ in regions],
                    [i.name for i in regions],
                )


class OperationAdator(ir.OpAdaptor):
    @classmethod
    def __init_subclass__(cls, *, operation: type):
        cls.OPERATION_NAME = operation.OPERATION_NAME
        cls._operation_cls = operation

        operands, attrs, results, regions = partition_fields(operation._fields)

        for attr in attrs:
            setattr(
                cls,
                attr.name,
                property(lambda self, name=attr.name: self.attributes[name]),
            )

        for i, operand in enumerate(operands):
            if operation._ODS_OPERAND_SEGMENTS:

                def getter(self, i=i, operand=operand):
                    operand_range = segmented_accessor(
                        self.operands,
                        self.attributes["operandSegmentSizes"],
                        i,
                    )
                    return normalize_value_range(operand_range, operand.variadicity)

                setattr(cls, operand.name, property(getter))
            else:
                setattr(cls, operand.name, property(lambda self, i=i: self.operands[i]))


@dataclass
class ParamDef:
    name: str
    constraint: Any


class Type(ir.DynamicType):
    """
    Base class of Python-defined types.

    The following example shows two ways to define types via this class:
    ```python
    class MyType(MyDialect.Type, name=..):
      ...

    class MyType(Type, dialect=MyDialect, name=..):
      ...
    ```
    """

    @classmethod
    def __init_subclass__(
        cls,
        *,
        name: str | None = None,
        dialect: type | None = None,
        **kwargs,
    ):
        super().__init_subclass__(**kwargs)

        fields = []

        for base in cls.__bases__:
            if hasattr(base, "_fields"):
                fields.extend(base._fields)
        for key, value in cls.__annotations__.items():
            field = ParamDef(key, value)
            fields.append(field)

        cls._fields = fields

        if dialect:
            if hasattr(cls, "_dialect_obj"):
                raise RuntimeError(
                    f"This type has already been attached to dialect '{cls._dialect_obj.DIALECT_NAMESPACE}'."
                )
            cls._dialect_obj = dialect

        # for subclasses without "name" parameter,
        # just treat them as normal classes
        if not name:
            return

        if not hasattr(cls, "_dialect_obj"):
            raise RuntimeError(
                "Type subclasses must either inherit from a Dialect's Type subclass "
                "or provide the dialect as a class keyword argument."
            )

        cls._name = name
        cls._dialect_name = cls._dialect_obj.DIALECT_NAMESPACE
        cls.type_name = f"{cls._dialect_name}.{name}"

        for i, field in enumerate(cls._fields):
            setattr(
                cls,
                field.name,
                property(lambda self, i=i: self.params[i]),
            )

        cls._dialect_obj.types.append(cls)

    @classmethod
    def get(cls, *args, context=None):
        args = [
            ir.TypeAttr.get(arg, context) if isinstance(arg, ir.Type) else arg
            for arg in args
        ]
        return cls(ir.DynamicType.get(cls.type_name, args, context=context))

    @classmethod
    def _emit_type(cls) -> None:
        ctx = ConstraintLoweringContext()

        t = irdl.type_(cls._name)
        with ir.InsertionPoint(t.body):
            irdl.parameters(
                [ctx.lower(f.constraint) for f in cls._fields],
                [f.name for f in cls._fields],
            )


class Attribute(ir.DynamicAttr):
    """
    Base class of Python-defined attributes.

    The following example shows two ways to define attributes via this class:
    ```python
    class MyAttr(MyDialect.Attribute, name=..):
      ...

    class MyAttr(Attribute, dialect=MyDialect, name=..):
      ...
    ```
    """

    @classmethod
    def __init_subclass__(
        cls,
        *,
        name: str | None = None,
        dialect: type | None = None,
        **kwargs,
    ):
        super().__init_subclass__(**kwargs)

        fields = []

        for base in cls.__bases__:
            if hasattr(base, "_fields"):
                fields.extend(base._fields)
        for key, value in cls.__annotations__.items():
            field = ParamDef(key, value)
            fields.append(field)

        cls._fields = fields

        if dialect:
            if hasattr(cls, "_dialect_obj"):
                raise RuntimeError(
                    f"This attribute has already been attached to dialect '{cls._dialect_obj.DIALECT_NAMESPACE}'."
                )
            cls._dialect_obj = dialect

        # for subclasses without "name" parameter,
        # just treat them as normal classes
        if not name:
            return

        if not hasattr(cls, "_dialect_obj"):
            raise RuntimeError(
                "Attribute subclasses must either inherit from a Dialect's Attribute subclass "
                "or provide the dialect as a class keyword argument."
            )

        cls._name = name
        cls._dialect_name = cls._dialect_obj.DIALECT_NAMESPACE
        cls.attr_name = f"{cls._dialect_name}.{name}"

        for i, field in enumerate(cls._fields):
            setattr(
                cls,
                field.name,
                property(lambda self, i=i: self.params[i]),
            )

        cls._dialect_obj.attributes.append(cls)

    @classmethod
    def get(cls, *args, context=None):
        args = [
            ir.TypeAttr.get(arg, context) if isinstance(arg, ir.Type) else arg
            for arg in args
        ]
        return cls(ir.DynamicAttr.get(cls.attr_name, args, context=context))

    @classmethod
    def _emit_attr(cls) -> None:
        ctx = ConstraintLoweringContext()

        t = irdl.attribute(cls._name)
        with ir.InsertionPoint(t.body):
            irdl.parameters(
                [ctx.lower(f.constraint) for f in cls._fields],
                [f.name for f in cls._fields],
            )


class Dialect(ir.Dialect):
    """
    Base class of a Python-defined dialect.

    It can be used like the following example:
    ```python
    class MyInt(Dialect, name="myint"):
        pass

    i32 = IntegerType[32]

    class ConstantOp(MyInt.Operation, name="constant"):
        value: IntegerAttr
        cst: Result[i32] = infer_result()

    class AddOp(MyInt.Operation, name="add"):
        lhs: Operand[i32]
        rhs: Operand[i32]
        res: Result[i32] = infer_result()
    ```
    """

    @classmethod
    def __init_subclass__(cls, name: str, **kwargs):
        cls.name = name
        cls.DIALECT_NAMESPACE = name
        cls.operations = []
        cls.Operation = type(
            "Operation",
            (Operation,),
            dict(),
            dialect=cls,
        )
        cls.types = []
        cls.Type = type(
            "Type",
            (Type,),
            dict(),
            dialect=cls,
        )
        cls.attributes = []
        cls.Attribute = type(
            "Attribute",
            (Attribute,),
            dict(),
            dialect=cls,
        )

    @classmethod
    def _emit_dialect(cls) -> None:
        d = irdl.dialect(cls.name)
        with ir.InsertionPoint(d.body):
            for type_ in cls.types:
                type_._emit_type()
            for attr in cls.attributes:
                attr._emit_attr()
            for op in cls.operations:
                op._emit_operation()

    @classmethod
    def _emit_module(cls) -> ir.Module:
        with ir.Location.unknown() if not ir.Location.current else nullcontext():
            m = ir.Module.create()
            with ir.InsertionPoint(m.body):
                cls._emit_dialect()

        return m

    @classmethod
    def load(
        cls,
        *,
        reload: bool = False,
    ) -> None:
        if hasattr(cls, "_mlir_module") and not reload:
            if cls._mlir_module.context is not ir.Context.current:
                raise RuntimeError(
                    "This dialect was loaded in a different context. "
                    "Please set reload=True to reload the dialect in the current context."
                )
            return

        cls._mlir_module = cls._emit_module()
        pm = PassManager()
        pm.add("canonicalize, cse")
        pm.run(cls._mlir_module.operation)

        irdl.load_dialects(cls._mlir_module)

        for op in cls.operations:
            op._attach_traits()

        _cext.globals._register_dialect_impl(cls.DIALECT_NAMESPACE, cls, replace=reload)

        for type_ in cls.types:
            typeid = ir.DynamicType.lookup_typeid(type_.type_name)
            _cext.register_type_caster(typeid, replace=reload)(type_)

        for attr in cls.attributes:
            typeid = ir.DynamicAttr.lookup_typeid(attr.attr_name)
            _cext.register_type_caster(typeid, replace=reload)(attr)

        for op in cls.operations:
            _cext.register_operation(cls, replace=reload)(op)
            _cext.register_op_adaptor(op, replace=reload)(op.Adaptor)



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from dataclasses import dataclass
import sys

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
from numba_cuda_mlir.tools import get_gpu_compute_capability, format_arch
from numba_cuda_mlir.numba_cuda.cudadrv.driver import _Linker


def _link_item_has_callbacks(link_item) -> bool:
    return bool(getattr(link_item, "setup_callback", None)) or bool(
        getattr(link_item, "teardown_callback", None)
    )


def _link_item_is_ltoir(link_item) -> bool:
    if isinstance(link_item, str):
        return link_item.endswith(".ltoir")
    return type(link_item).__name__ == "LTOIR"


def _link_item_is_cuda_source(link_item) -> bool:
    if isinstance(link_item, str):
        return link_item.endswith(".cu")
    return type(link_item).__name__ == "CUSource"


@dataclass(frozen=True)
class ResolvedLinkPlan:
    """Resolved JIT link policy.

    ``compile_new_inputs_as_ltoir`` controls newly generated inputs: the current
    MLIR module and CUDA source link items. Existing LTOIR link inputs still
    make the CUDA linker use LTO even when this flag is false.
    """

    compile_new_inputs_as_ltoir: bool
    lto_explicit: bool
    requested_lto: bool
    compile_output: str
    has_external_link_items: bool
    has_ltoir_link_items: bool
    has_callback_link_items: bool


def resolve_link_plan(targetoptions: dict, link_items=(), extra_ltoir_items=()) -> ResolvedLinkPlan:
    link_items = tuple(link_items or ())
    extra_ltoir_items = tuple(extra_ltoir_items or ())
    lto_explicit = bool(targetoptions.get("_lto_explicit", False))
    requested_lto = bool(targetoptions.get("lto", False))
    compile_output = targetoptions.get("_compile_output", "ptx")
    has_external_link_items = bool(link_items)
    has_ltoir_link_items = bool(extra_ltoir_items) or any(
        _link_item_is_ltoir(link_item) for link_item in link_items
    )
    has_ptx_link_items = any(
        link_item.endswith(".ptx")
        if isinstance(link_item, str)
        else type(link_item).__name__ == "PTXSource"
        for link_item in link_items
    )
    has_cuda_source_link_items = any(
        _link_item_is_cuda_source(link_item) for link_item in link_items
    )
    has_callback_link_items = any(_link_item_has_callbacks(link_item) for link_item in link_items)

    if compile_output not in ("ptx", "ltoir"):
        raise ValueError(f"Unsupported compile output: {compile_output}")

    if has_ltoir_link_items and lto_explicit and not requested_lto:
        raise ValueError("Cannot link LTOIR inputs with lto=False")

    if compile_output == "ltoir":
        compile_new_inputs_as_ltoir = True
    elif lto_explicit:
        compile_new_inputs_as_ltoir = requested_lto
    elif requested_lto:
        compile_new_inputs_as_ltoir = True
    elif has_ptx_link_items or targetoptions.get("debug", False):
        compile_new_inputs_as_ltoir = False
    else:
        compile_new_inputs_as_ltoir = has_ltoir_link_items or has_cuda_source_link_items

    return ResolvedLinkPlan(
        compile_new_inputs_as_ltoir=compile_new_inputs_as_ltoir,
        lto_explicit=lto_explicit,
        requested_lto=requested_lto,
        compile_output=compile_output,
        has_external_link_items=has_external_link_items,
        has_ltoir_link_items=has_ltoir_link_items,
        has_callback_link_items=has_callback_link_items,
    )


class Linker(_Linker):
    def __init__(
        self,
        cc: tuple[int, int] | None = None,
        additional_flags: list[str] | None = None,
        lto: bool | None = None,
        arch: str | None = None,
        verbose: bool = False,
        ftz: bool | None = None,
        prec_div: bool | None = None,
        prec_sqrt: bool | None = None,
        fma: bool | None = None,
        debug: bool = False,
        lineinfo: bool = False,
        optimize_unused_variables: bool = True,
        variables_used: str | tuple[str, ...] | list[str] | None = None,
        optimization_level: int = 3,
        ptxas_options: str | None = None,
        max_registers: int | None = None,
    ):
        if cc is None:
            cc = get_gpu_compute_capability(tuple)
        if arch is None:
            arch = format_arch(cc)
        super().__init__(
            cc=cc,
            additional_flags=additional_flags,
            lineinfo=lineinfo,
            lto=lto,
            max_registers=max_registers,
            arch=arch,
            verbose=verbose,
            debug=debug,
            ftz=ftz,
            prec_div=prec_div,
            prec_sqrt=prec_sqrt,
            fma=fma,
            optimize_unused_variables=optimize_unused_variables,
            variables_used=variables_used,
            optimization_level=optimization_level,
            ptxas_options=ptxas_options,
        )

        self._numba_cuda_mlir_temp_ptx_files: list[str] = []
        self._ltoirs: dict[int, bytes] = {}

    def add_file_guess_ext(
        self, path_or_code, ignore_nonlto=False, compile_cu_as_ltoir: bool | None = None
    ):
        if compile_cu_as_ltoir is None or not _link_item_is_cuda_source(path_or_code):
            return super().add_file_guess_ext(path_or_code, ignore_nonlto=ignore_nonlto)

        old_lto = self.lto
        self.lto = compile_cu_as_ltoir
        try:
            return super().add_file_guess_ext(path_or_code, ignore_nonlto=ignore_nonlto)
        finally:
            self.lto = old_lto

    def recreate_with_lto(
        self,
        lto: bool = True,
        ltoir_only: bool = False,
        kernel_ltoir: bytes | None = None,
    ) -> Self:
        """Recreate the linker, re-adding all object codes from raw bytes.

        When *ltoir_only* is True, only LTOIR objects are copied (useful for
        diagnostic ``-ptx`` links that require all inputs to be LTOIR).

        When *kernel_ltoir* is provided, it is added *first* in the link chain,
        ahead of any previously accumulated device-function and library
        LTO-IRs, so that nvJitLink resolves symbols against the kernel first.
        """
        self._materialize_pending_cu()

        existing = list(getattr(self, "_object_codes", []))
        new_linker = Linker(
            cc=self.cc,
            additional_flags=self.additional_flags,
            lto=lto,
            arch=self.arch,
            verbose=self._verbose,
            ftz=self._ftz,
            prec_div=self._prec_div,
            prec_sqrt=self._prec_sqrt,
            fma=self._fma,
            debug=self._debug,
            lineinfo=self.lineinfo,
            optimize_unused_variables=self._optimize_unused_variables,
            variables_used=self.variables_used,
            optimization_level=self._optimization_level,
            ptxas_options=self._ptxas_options,
            max_registers=self.max_registers,
        )
        if kernel_ltoir is not None:
            new_linker.add_ltoir(kernel_ltoir)
        for obj in existing:
            code_type = getattr(obj, "code_type", None)
            if ltoir_only and code_type != "ltoir":
                continue
            if code_type == "ptx":
                new_linker.add_ptx(obj.code)
            elif code_type == "cubin":
                new_linker.add_cubin(obj.code)
            elif code_type == "ltoir":
                new_linker.add_ltoir(obj.code)
            else:
                new_linker._object_codes.append(obj)

        # Preserve dedup tracking from the source linker without dropping the
        # kernel LTO-IR that may have been prepended above.
        for h, ltoir in self._ltoirs.items():
            new_linker._ltoirs.setdefault(h, ltoir)
        return new_linker

    def add_ltoir(self, ltoir: bytes, name: str = "") -> None:
        h = hash(ltoir)
        if h in self._ltoirs:
            return self
        if not self.lto:
            new_linker = self.recreate_with_lto()
            self.__dict__.update(new_linker.__dict__)
        super().add_ltoir(ltoir, name)
        self._ltoirs[h] = ltoir

    def merge_ltoirs_from(self, other: "Linker") -> None:
        for ltoir in other._ltoirs.values():
            self.add_ltoir(ltoir)

    def _get_linker_options(self, ptx):
        return super()._get_linker_options(ptx)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import re
from functools import cached_property
import warnings
import importlib.util
import numpy as np

from numba_cuda_mlir.numba_cuda import types
from numba_cuda_mlir.numba_cuda import HAS_NUMBA
from numba_cuda_mlir.numba_cuda.core.callconv import CUDACallConv
from numba_cuda_mlir.numba_cuda.core.compiler_lock import global_compiler_lock
from numba_cuda_mlir.numba_cuda.core.errors import NumbaWarning
from numba_cuda_mlir.numba_cuda.core.base import BaseContext
from numba_cuda_mlir.numba_cuda.typing import cmathdecl
from numba_cuda_mlir.numba_cuda import datamodel

from .cudadrv import nvvm
from numba_cuda_mlir.numba_cuda import (
    cgutils,
    itanium_mangler,
    compiler,
    codegen,
    typing,
)
from numba_cuda_mlir.numba_cuda.debuginfo import CUDADIBuilder
from numba_cuda_mlir.numba_cuda.flags import CUDAFlags
from numba_cuda_mlir.numba_cuda.models import cuda_data_manager
from numba_cuda_mlir.numba_cuda.core import config, targetconfig


# -----------------------------------------------------------------------------
# Typing


class CUDATypingContext(typing.BaseContext):
    def load_additional_registries(self):
        from . import (
            cudadecl,
            cudamath,
            fp16,
            bf16,
            fp8,
            libdevicedecl,
            vector_types,
        )
        from numba_cuda_mlir.numba_cuda.typing import enumdecl, cffi_utils, npydecl
        from numba_cuda_mlir.extending import (
            typing_registry as extending_typing_registry,
        )

        self.install_registry(cudadecl.registry)
        self.install_registry(cffi_utils.registry)
        self.install_registry(cudamath.registry)
        self.install_registry(cmathdecl.registry)
        self.install_registry(libdevicedecl.registry)
        self.install_registry(npydecl.registry)
        self.install_registry(enumdecl.registry)
        self.install_registry(vector_types.typing_registry)
        self.install_registry(fp16.typing_registry)
        self.install_registry(bf16.typing_registry)
        self.install_registry(fp8.typing_registry)
        self.install_registry(extending_typing_registry)

    def resolve_value_type(self, val):
        # treat other dispatcher object as another device function
        from numba_cuda_mlir.numba_cuda.dispatcher import CUDADispatcher
        from numba_cuda_mlir.numba_cuda.dispatcher import Dispatcher

        if HAS_NUMBA:
            if isinstance(val, Dispatcher) and not isinstance(val, CUDADispatcher):
                try:
                    # use cached device function
                    val = val.__dispatcher
                except AttributeError:
                    if not val._can_compile:
                        raise ValueError(
                            "using cpu function on device but its compilation is disabled"
                        )
                    targetoptions = val.targetoptions.copy()
                    targetoptions["device"] = True
                    targetoptions["debug"] = targetoptions.get("debug", False)
                    targetoptions["opt"] = targetoptions.get("opt", True)
                    disp = CUDADispatcher(val.py_func, targetoptions)
                    # cache the device function for future use and to avoid
                    # duplicated copy of the same function.
                    val.__dispatcher = disp
                    val = disp

        # continue with parent logic
        return super().resolve_value_type(val)

    def can_convert(self, fromty, toty):
        """
        Check whether conversion is possible from *fromty* to *toty*.
        If successful, return a numba.cuda.typeconv.Conversion instance;
        otherwise None is returned.
        """

        # This implementation works around the issue addressed in Numba PR
        # #10047, "Fix IntEnumMember.can_convert_to() when no conversions
        # found", https://github.com/numba/numba/pull/10047.
        #
        # This should be gated on the version of Numba that the fix is
        # incorporated into, and eventually removed when the minimum supported
        # Numba version includes the fix.

        try:
            return super().can_convert(fromty, toty)
        except TypeError:
            if isinstance(fromty, types.IntEnumMember):
                # IntEnumMember fails to correctly handle impossible
                # conversions - in this scenario the correct thing to do is to
                # return None to signal that the conversion was not possible
                return None
            else:
                # Any failure involving conversion from a non-IntEnumMember is
                # almost certainly a real and separate issue
                raise


# -----------------------------------------------------------------------------
# Implementation


VALID_CHARS = re.compile(r"[^a-z0-9]", re.I)


def load_cuda_target_registration_modules():
    from numba_cuda_mlir.numba_cuda.cpython import (
        numbers,
        slicing,
        iterators,
        unicode,
        charseq,
        cmathimpl,
        mathimpl,
        tupleobj,
        rangeobj,
        enumimpl,
    )
    from numba_cuda_mlir.numba_cuda.cpython import builtins as cpython_builtins
    from numba_cuda_mlir.numba_cuda.core import optional
    from numba_cuda_mlir.numba_cuda.misc import cffiimpl
    from numba_cuda_mlir.numba_cuda.np import (
        arrayobj,
        npdatetime,
        polynomial,
        arraymath,
    )
    from numba_cuda_mlir.numba_cuda.np.unsafe import ndarray

    return (
        numbers,
        slicing,
        iterators,
        unicode,
        charseq,
        cmathimpl,
        mathimpl,
        tupleobj,
        rangeobj,
        enumimpl,
        cpython_builtins,
        optional,
        cffiimpl,
        arrayobj,
        npdatetime,
        polynomial,
        arraymath,
        ndarray,
    )


class CUDATargetContext(BaseContext):
    implement_powi_as_math_call = True
    strict_alignment = True

    def __init__(self, typingctx, target="cuda"):
        super().__init__(typingctx, target)
        self.data_model_manager = cuda_data_manager.chain(datamodel.default_manager)

    @property
    def enable_nrt(self):
        return getattr(config, "CUDA_ENABLE_NRT", False)

    @property
    def DIBuilder(self):
        return CUDADIBuilder

    @property
    def enable_boundscheck(self):
        # Unconditionally disabled
        return False

    # Overrides
    def create_module(self, name):
        return self._internal_codegen._create_empty_module(name)

    def init(self):
        self._internal_codegen = codegen.JITCUDACodegen("numba.cuda.jit")
        self._target_data = None

    def load_additional_registries(self):
        (
            numbers,
            slicing,
            iterators,
            unicode,
            charseq,
            cmathimpl,
            mathimpl,
            tupleobj,
            rangeobj,
            enumimpl,
            cpython_builtins,
            optional,
            cffiimpl,
            arrayobj,
            npdatetime,
            polynomial,
            arraymath,
            _,
        ) = load_cuda_target_registration_modules()

        from . import (
            cudaimpl,
            fp16,
            mathimpl as cuda_mathimpl,
            vector_types,
            bf16,
            fp8,
        )

        self.install_registry(cudaimpl.registry)
        self.install_registry(cffiimpl.registry)
        self.install_registry(cmathimpl.registry)
        self.install_registry(mathimpl.registry)
        self.install_registry(numbers.registry)
        self.install_registry(optional.registry)
        self.install_registry(cuda_mathimpl.registry)
        self.install_registry(vector_types.impl_registry)
        self.install_registry(fp16.target_registry)
        self.install_registry(bf16.target_registry)
        self.install_registry(fp8.target_registry)
        self.install_registry(slicing.registry)
        self.install_registry(iterators.registry)
        self.install_registry(unicode.registry)
        self.install_registry(charseq.registry)
        self.install_registry(tupleobj.registry)
        self.install_registry(rangeobj.registry)
        self.install_registry(enumimpl.registry)
        self.install_registry(cpython_builtins.registry)

        # install np registries
        self.install_registry(polynomial.registry)
        self.install_registry(npdatetime.registry)
        self.install_registry(arrayobj.registry)
        self.install_registry(arraymath.registry)

    def codegen(self):
        return self._internal_codegen

    @property
    def target_data(self):
        # The MLIR pipeline does not use an llvmlite TargetData; this property
        # is only reachable on the dead llvmlite codegen path.
        raise NotImplementedError("CUDATargetContext.target_data is not available on the MLIR path")

    @cached_property
    def nonconst_module_attrs(self):
        """
        Some CUDA intrinsics are at the module level, but cannot be treated as
        constants, because they are loaded from a special register in the PTX.
        These include threadIdx, blockDim, etc.
        """
        from numba_cuda_mlir import numba_cuda as cuda

        nonconsts = (
            "threadIdx",
            "blockDim",
            "blockIdx",
            "gridDim",
            "laneid",
            "warpsize",
        )
        nonconsts_with_mod = tuple([(types.Module(cuda), nc) for nc in nonconsts])
        return nonconsts_with_mod

    @property
    def call_conv(self):
        warnings.warn(
            "Context.call_conv is deprecated. Use FunctionDescriptor.call_conv instead.",
            DeprecationWarning,
        )
        return self.fndesc.call_conv

    def make_constant_array(self, builder, aryty, arr):
        # Builds an llvmlite constant array global; only reachable on the dead
        # llvmlite codegen path. The MLIR pipeline materializes constant arrays
        # itself, so this is never called.
        raise NotImplementedError(
            "CUDATargetContext.make_constant_array is not available on the MLIR path"
        )

    def insert_const_string(self, mod, string):
        # Builds an llvmlite constant string global; only reachable on the dead
        # llvmlite codegen path.
        raise NotImplementedError(
            "CUDATargetContext.insert_const_string is not available on the MLIR path"
        )

    def insert_string_const_addrspace(self, builder, string):
        # Builds an llvmlite constant string global; only reachable on the dead
        # llvmlite codegen path.
        raise NotImplementedError(
            "CUDATargetContext.insert_string_const_addrspace is not available on the MLIR path"
        )

    def optimize_function(self, func):
        """Run O1 function passes"""
        pass
        ## XXX skipped for now
        # fpm = lp.FunctionPassManager.new(func.module)
        #
        # lp.PassManagerBuilder.new().populate(fpm)
        #
        # fpm.initialize()
        # fpm.run(func)
        # fpm.finalize()

    def get_ufunc_info(self, ufunc_key):
        # The ufunc loop -> llvmlite codegen table (numba_cuda.ufuncs) only fed
        # the vendored numpy ufunc lowering (_KernelImpl), which is filtered out
        # on the MLIR path. Ufunc loop *typing* uses np.ufunc_db instead, and
        # ufuncs are lowered by numba_cuda_mlir.lowering.numpy. Reaching here
        # means dead vendored ufunc codegen ran.
        raise NotImplementedError(
            "CUDATargetContext.get_ufunc_info is not available on the MLIR path"
        )

    def _compile_subroutine_no_cache(self, builder, impl, sig, locals=None, flags=None):
        # Overrides numba.core.base.BaseContext._compile_subroutine_no_cache().
        # Modified to use flags from the context stack if they are not provided
        # (pending a fix in Numba upstream).

        if locals is None:
            locals = {}

        with global_compiler_lock:
            codegen = self.codegen()
            library = codegen.create_library(impl.__name__)
            if flags is None:
                cstk = targetconfig.ConfigStack()
                if cstk:
                    flags = cstk.top().copy()
                else:
                    msg = "There should always be a context stack; none found."
                    warnings.warn(msg, NumbaWarning)
                    flags = CUDAFlags()

            flags.no_compile = True
            flags.no_cpython_wrapper = True
            flags.no_cfunc_wrapper = True

            # compile_subroutine always uses CUDACallConv
            call_conv = CUDACallConv(self)
            abi_info = {}

            cres = compiler.compile_internal(
                self.typing_context,
                self,
                library,
                impl,
                sig.args,
                sig.return_type,
                flags,
                locals=locals,
                call_conv=call_conv,
                abi_info=abi_info,
            )

            # Allow inlining the function inside callers
            self.active_code_library.add_linking_library(cres.library)
            return cres



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from numba_cuda_mlir.numba_cuda.cudadrv import devices, driver


def ndarray_device_allocate_data(ary):
    """
    Allocate gpu data buffer
    """
    datasize = driver.host_memory_size(ary)
    # allocate
    gpu_data = devices.get_context().memalloc(datasize)
    return gpu_data



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from numba_cuda_mlir._mlir.dialects.llvm import *  # noqa: F401,F403
from numba_cuda_mlir._mlir.dialects import llvm as _llvm
from numba_cuda_mlir._mlir import ir


def ptr():
    """Shorthand for ``llvm.PointerType.get()``."""
    return _llvm.PointerType.get()


def insertvalue(container, value, position, **kwargs):
    """Like ``llvm.insertvalue`` but *position* can be a plain ``int`` or ``list[int]``."""
    if isinstance(position, int):
        position = ir.DenseI64ArrayAttr.get([position])
    elif isinstance(position, (list, tuple)):
        position = ir.DenseI64ArrayAttr.get(list(position))
    return _llvm.insertvalue(container=container, value=value, position=position, **kwargs)


def addressof(global_name, *, res=None, loc=None, ip=None):
    """Return a pointer to the named global.

    *global_name* can be a plain ``str`` or a ``FlatSymbolRefAttr``.
    *res* defaults to ``!llvm.ptr`` if not given.
    """
    if res is None:
        res = ptr()
    return _llvm.mlir_addressof(res, global_name, loc=loc, ip=ip)



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# CUDA-callables that are implemented in MLIR

import sys
import inspect
import numba_cuda_mlir.runtime as rt
from numba_cuda_mlir.compiler import declare_mlir_library
from numba_cuda_mlir.errors import (
    MultipleIntrinsicFunctionsError,
    UnsupportedIntrinsicTypeError,
)
from numba_cuda_mlir.lowering_utilities.type_conversions import to_mlir_type
from numba_cuda_mlir.lowering_utilities.discover_functions import discover_functions
from numba_cuda_mlir._mlir.ir import UnitAttr


def _import_runtime():
    for lib in rt._get_all_libraries():
        lib = getattr(rt, lib)
        for func in lib.functions.keys():
            setattr(sys.modules[__name__], func, getattr(lib, func))


_import_runtime()

breakpoint = rt.util.breakpoint
nanosleep = rt.util.nanosleep


def _resolve_type_annotation(ty):
    if ty is None:
        return None  # void return type
    # Handle tuple return types (e.g., tuple[types.int32, types.int32])
    if hasattr(ty, "__origin__") and ty.__origin__ is tuple:
        return [to_mlir_type(arg) for arg in ty.__args__]
    try:
        return to_mlir_type(ty)
    except (TypeError, NotImplementedError):
        pass
    if callable(ty):
        return ty()
    raise TypeError(f"Cannot resolve type annotation: {ty}")


def _intrinsic_from_source(source: str):
    try:
        functions = discover_functions(source)
    except (TypeError, NotImplementedError) as e:
        raise UnsupportedIntrinsicTypeError(str(e)) from e

    if len(functions) == 0:
        raise ValueError(
            "No functions found in MLIR source. Ensure the source contains a func.func definition."
        )

    if len(functions) > 1:
        raise MultipleIntrinsicFunctionsError(list(functions.keys()))

    func_name = next(iter(functions.keys()))
    lib = declare_mlir_library(source)
    return getattr(lib, func_name)


def _intrinsic_from_func(func):
    from numba_cuda_mlir.mlir.context import mlir_mod_ctx
    from numba_cuda_mlir.mlir.dialect_exts.func import func as mlir_func

    sig = inspect.signature(func)

    original_annotations = {}
    for name, param in sig.parameters.items():
        ty = param.annotation
        if ty is inspect.Parameter.empty:
            raise TypeError(f"Parameter '{name}' must have a type annotation")
        original_annotations[name] = ty

    ret_ty = sig.return_annotation
    if ret_ty is not inspect.Signature.empty and ret_ty is not None:
        original_annotations["return"] = ret_ty

    with mlir_mod_ctx() as ctx:
        new_annotations = {}
        for name, ty in original_annotations.items():
            resolved = _resolve_type_annotation(ty)
            if resolved is not None:
                new_annotations[name] = resolved

        func.__annotations__ = new_annotations

        decorator = mlir_func(sym_visibility="private")
        mlir_fn = decorator(func)
        mlir_fn.func_attrs["alwaysinline"] = UnitAttr.get()
        mlir_fn.emit()

        # Capture the module string before exiting the context
        module_str = str(ctx.module)

    # Restore original annotations to avoid retaining references to MLIR
    # objects after the context is destroyed (use-after-free)
    func.__annotations__ = original_annotations

    lib = declare_mlir_library(module_str)
    return getattr(lib, func.__name__)


def define(func_or_source):
    """
    Create an intrinsic function callable from JIT kernels.

    Can be used in two ways:

    1. As a decorator for Python functions using MLIR Python bindings:

        @cuda.intrin.define
        def my_mul(a: types.int32, b: types.int32) -> types.int32:
            return arith.muli(a, b)

    2. With inline MLIR source code:

        my_add = cuda.intrin.define('''
            func.func private @my_add(%a: i32, %b: i32) -> i32 attributes {always_inline} {
                %r = arith.addi %a, %b : i32
                return %r : i32
            }
        ''')

    For the decorator form, type annotations can be:
    - numba_cuda_mlir types: `types.int32`, `types.float32`, etc. (recommended)
    - MLIR type callables: `T.f32`, `llvm.PointerType.get` (for advanced types)

    Args:
        func_or_source: Either a Python function with type annotations, or
                        MLIR source code containing exactly one func.func.

    Returns:
        An ExternMLIRLibraryFunction that can be called from JIT kernels.

    Raises:
        TypeError: If used as decorator and a parameter lacks a type annotation.
        MultipleIntrinsicFunctionsError: If MLIR source contains multiple functions.
        UnsupportedIntrinsicTypeError: If a type cannot be converted.
        ValueError: If MLIR source contains no functions.
    """
    if isinstance(func_or_source, str):
        return _intrinsic_from_source(func_or_source)
    else:
        return _intrinsic_from_func(func_or_source)



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# Common utilities for AST transformations
import ast
import inspect
import textwrap
import types
from typing import Callable


def get_function_ast(func: Callable) -> ast.Module | None:
    """Parse and return the AST for a function, or None if source unavailable."""
    try:
        source = inspect.getsource(func)
        source = textwrap.dedent(source)
        return ast.parse(source)
    except (OSError, TypeError):
        return None


def get_function_context(func: Callable) -> dict:
    """Build a context dict from a function's globals and closure variables."""
    context = func.__globals__.copy()

    # Add closure variables (skip cells that are not yet assigned, e.g.
    # recursive self-references at decoration time).
    if func.__closure__ is not None:
        for name, cell in zip(func.__code__.co_freevars, func.__closure__):
            try:
                context[name] = cell.cell_contents
            except ValueError:
                pass

    return context


def recompile_function(func: Callable, tree: ast.Module, stored_values: dict = None) -> Callable:
    """Compile a modified AST back into a function object.

    Args:
        func: The original function
        tree: The modified AST
        stored_values: Dict of name -> value for complex objects that need to be
            injected into the function's globals
    """
    code = compile(tree, inspect.getfile(func), "exec")

    # Find the function's code object in the compiled module
    func_code = None
    for const in code.co_consts:
        if isinstance(const, types.CodeType) and const.co_name == func.__name__:
            func_code = const
            break

    if func_code is None:
        raise ValueError(f"Could not find compiled code for {func.__name__}")

    # Build closure for the new function - only include cells that are still needed
    new_closure = None
    if func_code.co_freevars and func.__closure__:
        old_freevars = func.__code__.co_freevars
        old_closure = dict(zip(old_freevars, func.__closure__))
        new_closure = tuple(old_closure[name] for name in func_code.co_freevars)

    # Create a new globals dict with closure variables and stored values injected
    new_globals = func.__globals__.copy()

    # Add closure variables to globals - the recompiled code may treat them as globals
    # since it was compiled from source without the original closure context.
    # Skip empty cells (e.g. recursive self-references at decoration time).
    if func.__closure__ is not None:
        for name, cell in zip(func.__code__.co_freevars, func.__closure__):
            try:
                new_globals[name] = cell.cell_contents
            except ValueError:
                pass

    if stored_values:
        new_globals.update(stored_values)

    # Create new function with the modified code
    new_func = types.FunctionType(
        func_code,
        new_globals,
        func.__name__,
        func.__defaults__,
        new_closure,
    )
    new_func.__annotations__ = func.__annotations__
    new_func.__doc__ = func.__doc__
    new_func.__module__ = func.__module__
    new_func.__qualname__ = func.__qualname__

    return new_func



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause


def initialize_all():
    # Import models to register them with the data model manager
    import numba_cuda_mlir.numba_cuda.models  # noqa: F401



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from numba_cuda_mlir.numba_cuda.types import (
    BaseTuple,
    BaseNamedTuple,
    BoundFunction,
    Optional,
    Poison,
    UnicodeType,
    UnionType,
    Boolean,
    PyObject,
    StringLiteral,
    BooleanLiteral,
    IntegerLiteral,
    Complex,
    NumberClass,
    NPDatetime,
    NPTimedelta,
    NumpyNdIterType,
    RangeType,
    RangeIteratorType,
    DType,
    DTypeSpec,
    VarArg,
    Any,
    Literal,
    string,
    Bytes,
    Function,
    Integer,
    Number,
    Float,
    Module,
    Type,
    NoneType,
    MemInfoPointer,
    float16,
    float32,
    float64,
    EnumMember,
    IntEnumMember,
    EnumClass,
    # String/char sequence types
    CharSeq,
    UnicodeCharSeq,
    # Numpy aliases
    b1,
    i8,
    u1,
    u2,
    u4,
    u8,
    f2,
    f4,
    f8,
    c8,
    c16,
    i1,
    i2,
    i4,
    int8,
    int16,
    int32,
    int64,
    uint8,
    uint16,
    uint32,
    uint64,
    complex32,
    complex64,
    complex128,
    bool,
    boolean,
    void,
    none,
    uintp,
    intp,
    voidptr,
    unicode_type,
    CPointer,
    RawPointer,
    Array,
    ArrayCTypes,
    TypeRef,
    Tuple,
    UniTuple,
    # Record/struct types
    Record,
    NestedArray,
    RecursiveCall,
    Omitted,
)
from numba_cuda_mlir.numba_cuda.typing import Signature, signature
from numba_cuda_mlir.type_defs.aggregate_types import AggregateType, UnionType
from numba_cuda_mlir.type_defs.cuda_types import CUTensorMapType, CUTensorMapStorageType
from numba_cuda_mlir.type_defs.pointer_types import ByValPointerType
from numba_cuda_mlir.type_defs.float_types import (
    BFloat16Type,
    NVFP4Type,
    Float4E2M1FNType,
    Float6E2M3FNType,
    Float6E3M2FNType,
    Float8E3M4Type,
    Float8E4M3B11FNUZType,
    Float8E4M3FNType,
    Float8E4M3FNUZType,
    Float8E4M3Type,
    Float8E5M2FNUZType,
    Float8E5M2Type,
    Float8E8M0FNUType,
    FloatTF32Type,
)


def get_numba_cuda_mlir_attributes(obj) -> dict:
    return getattr(obj, "__numba_cuda_mlir_attributes__", dict())


def set_numba_cuda_mlir_attributes(obj, attributes: dict):
    obj.__numba_cuda_mlir_attributes__ = attributes
    return obj


def grid_constant(type: Type) -> Type:
    attributes = get_numba_cuda_mlir_attributes(type)
    attributes["nvvm.grid_constant"] = True
    set_numba_cuda_mlir_attributes(type, attributes)
    return type


def alignas(type: Type, align: int) -> Type:
    attributes = get_numba_cuda_mlir_attributes(type)
    attributes["llvm.align"] = align
    set_numba_cuda_mlir_attributes(type, attributes)
    return type


i8 = int8
i16 = int16
i32 = int32
i64 = int64
u8 = uint8
u16 = uint16
u32 = uint32
u64 = uint64
f16 = float16
f32 = float32
f64 = float64

# TODO(ajm): alias the real types
bf16 = BFloat16Type()
nvfp4 = NVFP4Type("nvfp4", 4)

f4E2M1FN = Float4E2M1FNType("f4E2M1FN", 4)
f6E2M3FN = Float6E2M3FNType("f6E2M3FN", 6)
f6E3M2FN = Float6E3M2FNType("f6E3M2FN", 6)
f8E3M4 = Float8E3M4Type("f8E3M4", 8)
f8E4M3B11FNUZ = Float8E4M3B11FNUZType("f8E4M3B11FNUZ", 8)
f8E4M3FN = Float8E4M3FNType("f8E4M3FN", 8)
f8E4M3FNUZ = Float8E4M3FNUZType("f8E4M3FNUZ", 8)
f8E4M3 = Float8E4M3Type("f8E4M3", 8)
f8E5M2FNUZ = Float8E5M2FNUZType("f8E5M2FNUZ", 8)
f8E5M2 = Float8E5M2Type("f8E5M2", 8)
f8E8M0FNU = Float8E8M0FNUType("f8E8M0FNU", 8)
tf32 = FloatTF32Type("tf32", 18)
ptr = CPointer(void)
ptr_to = CPointer
byval_ptr = ByValPointerType

# CUDA-specific types
CUTensorMapStorage = CUTensorMapStorageType()
# we need a better representation of byval pointers in the type system.
# Probably we don't expose to users directly but we decay struct parameters
# to byval(struct)
CUTensorMap = alignas(byval_ptr(CUTensorMapStorage), 128)

_FP8_ATTRS = (
    "fp8_e5m2",
    "fp8_e4m3",
    "fp8_e8m0",
    "_type_fp8_e5m2",
    "_type_fp8_e4m3",
    "_type_fp8_e8m0",
    "cvt_e8m0_to_bf16raw",
    "bfloat16_raw_type",
    "bfloat16_raw_class",
)


def __getattr__(name):
    if name == "VectorTypeClass":
        from numba_cuda_mlir.typing.cuda_vector_types import VectorTypeClass

        return VectorTypeClass
    if name not in _FP8_ATTRS:
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}")
    import sys
    from numba_cuda_mlir.numba_cuda._internal import cuda_fp8

    val = (
        type(cuda_fp8.bfloat16_raw_type)
        if name == "bfloat16_raw_class"
        else getattr(cuda_fp8, name)
    )
    setattr(sys.modules[__name__], name, val)
    return val
